import dataclasses
import pathlib

from pyzipper import AESZipFile

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import (
    PocketSingerMetadata,
    PocketSingerNote,
    PocketSingerProject,
)
from .options import InputOptions


@dataclasses.dataclass
class PocketSingerParser:
    options: InputOptions
    archive_file: AESZipFile
    path: pathlib.Path
    first_bar_length: int = dataclasses.field(init=False)
    project: PocketSingerProject = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def parse_project(self, metadata: PocketSingerMetadata) -> Project:
        self.project = PocketSingerProject.model_validate_json(
            self.archive_file.read(metadata.ace_file_name)
        )
        song_tempo_list = self.parse_tempos()
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=self.parse_time_signatures(),
            track_list=self.parse_singing_tracks() + self.parse_instrumental_tracks(),
        )

    def parse_tempos(self) -> list[SongTempo]:
        return [SongTempo(bpm=self.project.song_info.bpm)]

    def parse_time_signatures(self) -> list[TimeSignature]:
        time_signatures = [
            TimeSignature(numerator=self.project.song_info.beat_of_bar, denominator=4)
        ]
        self.first_bar_length = round(time_signatures[0].bar_length())
        return time_signatures

    def parse_singing_tracks(self) -> list[SingingTrack]:
        singing_tracks: list[SingingTrack] = []
        for track in self.project.tracks:
            notes, pitch_points = self.parse_notes(track.notes)
            singing_track = SingingTrack(
                mute=track.mute,
                solo=track.solo,
                pan=track.pan,
                ai_singer_name=track.role_info.name,
                note_list=notes,
            )
            if self.options.import_pitch:
                pitch_points.insert(0, Point.start_point())
                pitch_points.append(Point.end_point())
                singing_track.edited_params.pitch.points.root = pitch_points
            singing_tracks.append(singing_track)
        return singing_tracks

    def parse_notes(self, ps_notes: list[PocketSingerNote]) -> tuple[list[Note], list[Point]]:
        notes: list[Note] = []
        pitch_points: list[Point] = []
        for ps_note in ps_notes:
            start_pos = int(self.synchronizer.get_actual_ticks_from_secs(ps_note.start_time))
            end_pos = int(self.synchronizer.get_actual_ticks_from_secs(ps_note.end_time))
            real_pitch = ps_note.pitch - 12 if self.project.version < 3 else ps_note.pitch
            note = Note(
                lyric=ps_note.grapheme if ps_note.grapheme_index == 0 else "+",
                key_number=real_pitch,
                start_pos=start_pos,
                length=end_pos - start_pos,
            )
            if ps_note.br_note is not None:
                note.head_tag = "V"
            if ps_note.consonant_time_head:
                note.edited_phones = Phones(head_length_in_secs=ps_note.consonant_time_head[0])
            if self.options.import_pitch:
                pitch_bends = ps_note.pitch_bends or ps_note.user_pitch or []
                if note_pitch_points := [
                    Point(
                        x=int(
                            self.synchronizer.get_actual_ticks_from_secs(pitch_bend.time)
                            + self.first_bar_length
                        ),
                        y=int((pitch_bend.pitch + real_pitch) * 100),
                    )
                    for pitch_bend in pitch_bends
                ]:
                    if note_pitch_points[-1].x < end_pos + self.first_bar_length:
                        note_pitch_points.append(
                            Point(x=end_pos + self.first_bar_length, y=note_pitch_points[-1].y)
                        )
                    note_pitch_points.insert(0, Point(x=note_pitch_points[0].x, y=-100))
                    note_pitch_points.append(Point(x=end_pos + self.first_bar_length, y=-100))
                    pitch_points.extend(note_pitch_points)
            notes.append(note)
        return notes, pitch_points

    def parse_instrumental_tracks(self) -> list[InstrumentalTrack]:
        instrumental_tracks: list[InstrumentalTrack] = []
        if self.options.import_instrumental_track and not hasattr(self.path, "protocol"):
            for bgm_track in self.project.bgm_info.tracks:
                audio_path = self.path.parent / f"{bgm_track.file_name}.{bgm_track.file_type}"
                if audio_path.exists():
                    instrumental_tracks.append(
                        InstrumentalTrack(
                            title=bgm_track.file_name,
                            mute=self.project.bgm_info.mute,
                            solo=self.project.bgm_info.solo,
                            audio_file_path=str(audio_path),
                            offset=self.synchronizer.get_actual_ticks_from_secs(
                                bgm_track.start_time
                            ),
                        )
                    )
        return instrumental_tracks



from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class ProcessOptions(BaseModel):
    key: int = Field(
        default=0,
        title=_("Key transition of pitch"),
        description=_("Pitch shift in semitones."),
    )



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .gjgj_generator import GjgjGenerator
from .gjgj_parser import GjgjParser
from .model import GjgjProject
from .options import InputOptions, OutputOptions


class GjgjConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        gjgj_project = GjgjProject.model_validate_json(path.read_bytes().decode("utf-8-sig"))
        return GjgjParser(options).parse_project(gjgj_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        gjgj_project = GjgjGenerator(options).generate_project(project)
        path.write_bytes(
            json.dumps(
                gjgj_project.model_dump(
                    mode="json",
                    exclude_none=True,
                    by_alias=True,
                ),
                ensure_ascii=False,
                separators=(",", ":"),
            ).encode("utf-8-sig")
        )



import dataclasses
import math
import random
from collections.abc import Callable

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.tick_counter import skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.audio import audio_track_info
from libresvip.utils.search import (
    binary_find_first,
    binary_find_last,
    find_last_index,
)

from .base_pitch_curve import BasePitchCurve
from .color_pool import count_color, get_color
from .enums import AcepLyricsLanguage
from .model import (
    AcepAudioPattern,
    AcepAudioTrack,
    AcepNote,
    AcepParamCurve,
    AcepParamCurveList,
    AcepParams,
    AcepProject,
    AcepSeedComposition,
    AcepSingerConfig,
    AcepTempo,
    AcepTimeSignature,
    AcepTrack,
    AcepVocalPattern,
    AcepVocalTrack,
)
from .options import OutputOptions, StrengthMappingOption
from .singers import DEFAULT_SEED, DEFAULT_SINGER_ID, singer2id, singer2seed


@dataclasses.dataclass
class AceGenerator:
    options: OutputOptions
    first_bar_tempo: list[SongTempo] = dataclasses.field(init=False)
    first_bar_ticks: int = dataclasses.field(init=False)
    ace_note_list: list[AcepNote] = dataclasses.field(init=False)
    pattern_start: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> AcepProject:
        ace_project = AcepProject()
        self.first_bar_ticks = int(project.time_signature_list[0].bar_length())
        self.first_bar_tempo = [
            tempo for tempo in project.song_tempo_list if tempo.position < self.first_bar_ticks
        ]
        denominator = project.time_signature_list[0].denominator
        numerator = project.time_signature_list[0].numerator
        ace_project.beats_per_bar = numerator * 4 // denominator
        if denominator <= 8 and numerator <= 8:
            ace_project.time_signatures = [
                AcepTimeSignature(bar_pos=0, numerator=numerator, denominator=denominator)
            ]
        self.synchronizer = TimeSynchronizer(project.song_tempo_list, self.first_bar_ticks)
        ace_project.tempos = self.generate_tempos(project.song_tempo_list)

        for track in project.track_list:
            if (ace_track := self.generate_track(track)) is not None:
                ace_project.tracks.append(ace_track)
        ace_project.duration = (
            max(
                (
                    len(ace_track)
                    for ace_track in ace_project.tracks
                    if isinstance(ace_track, AcepVocalTrack)
                ),
                default=0,
            )
        ) + 115200
        color_count = count_color()
        color_index = random.randint(0, color_count - 1)
        for ace_track in ace_project.tracks:
            ace_track.color = get_color(color_index)
            color_index = (color_index + 1) % color_count
        ace_project.color_index = color_index
        return ace_project

    def generate_tempos(self, tempos: list[SongTempo]) -> list[AcepTempo]:
        tempos = skip_tempo_list(tempos, self.first_bar_ticks)
        return [
            AcepTempo(
                bpm=tempo.bpm,
                position=tempo.position,
            )
            for tempo in tempos
        ]

    def generate_track(self, track: Track) -> AcepTrack | None:
        if isinstance(track, InstrumentalTrack):
            ace_audio_track = AcepAudioTrack()
            audio_pattern = AcepAudioPattern(path=track.audio_file_path)
            audio_pattern.pos = self.synchronizer.get_actual_secs_from_ticks(track.offset)
            if (track_info := audio_track_info(track.audio_file_path)) is not None:
                audio_pattern.dur = track_info.duration / 1000
                audio_pattern.clip_dur = audio_pattern.dur - audio_pattern.clip_pos
            else:
                return None
            ace_audio_track.patterns.append(audio_pattern)
            ace_track = ace_audio_track
        elif isinstance(track, SingingTrack):
            ace_vocal_track = AcepVocalTrack(
                language=self.options.lyric_language,
            )
            singer_config = AcepSingerConfig()
            if track.ai_singer_name in singer2id and track.ai_singer_name in singer2seed:
                singer_config.singer.singer_id = singer2id.get(
                    track.ai_singer_name, DEFAULT_SINGER_ID
                )
                singer_config.singer.composition.append(
                    AcepSeedComposition(code=singer2seed.get(track.ai_singer_name, DEFAULT_SEED))
                )
            else:
                singer_config.singer.singer_id = DEFAULT_SINGER_ID
                singer_config.singer.composition.append(AcepSeedComposition(code=DEFAULT_SEED))
            ace_vocal_track.singers.append(singer_config)
            if len(track.note_list):
                buffer = [track.note_list[0]]

                def generate_vocal_pattern() -> None:
                    self.pattern_start = round(
                        max(
                            0,
                            buffer[0].start_pos - 240,
                        )
                    )
                    self.ace_note_list = [self.generate_note(note) for note in buffer if note.lyric]
                    vocal_pattern = AcepVocalPattern(
                        pos=self.pattern_start,
                        dur=round(buffer[-1].end_pos) - self.pattern_start,
                        notes=self.ace_note_list,
                    )
                    vocal_pattern.clip_dur = vocal_pattern.dur
                    buffer.clear()
                    if self.options.breath > 0:
                        self.adjust_breath_tags(vocal_pattern.notes)
                    vocal_pattern.parameters = self.generate_params(track.edited_params)
                    ace_vocal_track.patterns.append(vocal_pattern)

                for prev_note, cur_note in zip(track.note_list[:-1], track.note_list[1:]):
                    prev_end = prev_note.end_pos
                    cur_start = cur_note.start_pos
                    if cur_start - prev_end > self.options.split_threshold * TICKS_IN_BEAT > 0:
                        generate_vocal_pattern()
                    buffer.append(cur_note)
                if buffer:
                    generate_vocal_pattern()
            ace_track = ace_vocal_track
        else:
            return None
        ace_track.name = track.title
        ace_track.mute = track.mute
        ace_track.solo = track.solo
        ace_track.pan = track.pan
        ace_track.gain = min(6.0, 20 * math.log10(track.volume))
        return ace_track

    def adjust_breath_tags(self, notes: list[AcepNote]) -> None:
        for i in range(1, len(notes)):
            breath = notes[i].br_len
            if breath == 0:
                continue
            distance = self.synchronizer.get_duration_secs_from_ticks(
                notes[i - 1].pos, notes[i].pos
            )
            actual_breath = min(distance / 2, breath)
            notes[i - 1].dur = min(
                notes[i - 1].dur,
                int(
                    self.synchronizer.get_actual_ticks_from_secs_offset(
                        notes[i - 1].pos, distance - actual_breath
                    )
                )
                - notes[i - 1].pos,
            )
            notes[i].br_len -= actual_breath

    def generate_note(self, note: Note) -> AcepNote:
        ace_note = AcepNote(
            pos=round(note.start_pos) - self.pattern_start,
            dur=note.length,
            pitch=note.key_number,
            lyric=note.lyric,
            language=self.options.lyric_language,
        )

        if all(symbol not in note.lyric for symbol in ["-", "+"]):
            if self.options.lyric_language == AcepLyricsLanguage.CHINESE:
                pronunciation = next(iter(get_pinyin_series(note.lyric)), None)
            else:
                pronunciation = None
            if note.pronunciation is not None and note.pronunciation != pronunciation:
                ace_note.syllable = note.pronunciation
            if note.edited_phones is not None and note.edited_phones.head_length_in_secs >= 0:
                ace_note.head_consonants = [note.edited_phones.head_length_in_secs]
            elif self.options.default_consonant_length:
                ace_note.head_consonants = [self.options.default_consonant_length]
        elif (
            self.options.lyric_language in [AcepLyricsLanguage.ENGLISH, AcepLyricsLanguage.SPANISH]
            and ace_note.lyric == "+"
            and self.ace_note_list
        ):
            ace_note.pronunciation = "-"
            last_ace_note = self.ace_note_list[
                find_last_index(self.ace_note_list, lambda n: n.lyric != "-")
            ]
            lyric, sep, index = last_ace_note.lyric.partition("#")
            if sep == "#" and index.isdigit():
                ace_note.lyric = f"{lyric}#{int(index) + 1}"
            else:
                last_ace_note.lyric = f"{lyric}#1"
                ace_note.lyric = f"{lyric}#2"
        else:
            ace_note.lyric = ace_note.pronunciation = "-"

        if note.head_tag == "V" and self.options.breath > 0:
            breath_start_in_secs = (
                self.synchronizer.get_actual_secs_from_ticks(note.start_pos)
                - self.options.breath / 1000
            )
            breath_start_in_ticks = self.synchronizer.get_actual_ticks_from_secs(
                breath_start_in_secs
            )
            ace_note.br_len = round(note.start_pos - breath_start_in_ticks)
        return ace_note

    @staticmethod
    def linear_transform(
        lower_bound: float, middle_value: float, upper_bound: float
    ) -> Callable[[float], float]:
        def transform(x: float) -> float:
            if x >= 0:
                return x * (upper_bound - middle_value) / 1000 + middle_value
            else:
                return x * (middle_value - lower_bound) / 1000 + middle_value

        return transform

    def generate_params(self, parameters: Params) -> AcepParams:
        result = AcepParams(
            breathiness=self.generate_param_curves(
                parameters.breath, self.linear_transform(0.2, 1, 2.5)
            ),
            gender=self.generate_param_curves(parameters.gender, self.linear_transform(-1, 0, 1)),
        )
        result.pitch_delta = self.generate_pitch_curves(parameters.pitch)
        if self.options.map_strength_info == StrengthMappingOption.BOTH:
            result.energy = self.generate_param_curves(
                parameters.strength,
                lambda x: self.linear_transform(0, 1, 2)(x / 2),
            )
            result.tension = self.generate_param_curves(
                parameters.strength,
                lambda x: self.linear_transform(0.7, 1, 1.5)(x / 2),
            )
        elif self.options.map_strength_info == StrengthMappingOption.ENERGY:
            result.energy = self.generate_param_curves(
                parameters.strength, self.linear_transform(0, 1, 2)
            )
        elif self.options.map_strength_info == StrengthMappingOption.TENSION:
            result.tension = self.generate_param_curves(
                parameters.strength, self.linear_transform(0.7, 1, 1.5)
            )
        return result

    def generate_pitch_curves(self, curve: ParamCurve) -> AcepParamCurveList:
        ace_curves = AcepParamCurveList()
        base_pitch = BasePitchCurve(self.ace_note_list, self.synchronizer, self.pattern_start)
        left_bound = max(0, self.pattern_start + self.ace_note_list[0].pos - 240)
        right_bound = max(
            0,
            self.pattern_start + self.ace_note_list[-1].pos + self.ace_note_list[-1].dur + 120,
        )

        segments = []
        for seg in curve.split_into_segments(-100):
            if seg[-1].x < self.first_bar_ticks:
                continue
            start_ticks = seg[0].x - self.first_bar_ticks if seg[0].x > self.first_bar_ticks else 0
            end_ticks = seg[-1].x - self.first_bar_ticks
            if start_ticks <= right_bound and end_ticks >= left_bound:
                segments.append(seg)

        for segment in segments:
            start_point = (
                binary_find_last(
                    segment,
                    lambda point: (0 <= point.x - self.first_bar_ticks <= left_bound),
                )
                or segment[0]
            )
            end_point = (
                binary_find_first(
                    segment,
                    lambda point: (right_bound <= point.x - self.first_bar_ticks),
                )
                or segment[-1]
            )
            if start_point.x == end_point.x:
                continue
            ace_curve = AcepParamCurve(
                offset=round(start_point.x - self.first_bar_ticks - self.pattern_start)
            )
            curve_end = round(end_point.x - self.first_bar_ticks - self.pattern_start)
            tick_step = (end_point.x - start_point.x) / (curve_end - ace_curve.offset)
            tick = float(start_point.x)
            while tick < self.first_bar_ticks:
                ace_curve.offset += 1
                tick += tick_step
            tick = max(self.first_bar_ticks, tick)
            tick_step = (end_point.x - tick) / (curve_end - ace_curve.offset)
            while tick < left_bound:
                ace_curve.offset += 1
                tick += tick_step
            pos = ace_curve.offset
            while pos <= min(right_bound, curve_end):
                second = self.synchronizer.get_actual_secs_from_ticks(
                    round(tick - self.first_bar_ticks)
                )
                ace_curve.values.append(
                    self.get_value_from_segment(segment, tick) / 100
                    - base_pitch.semitone_value_at(second)
                )
                pos += 1
                tick += tick_step
            ace_curves.root.append(ace_curve)
        return ace_curves

    def generate_param_curves(
        self, curve: ParamCurve, mapping_func: Callable[[float], float]
    ) -> AcepParamCurveList:
        ace_curves = AcepParamCurveList()
        left_bound = max(0, self.pattern_start + self.ace_note_list[0].pos - 240)
        right_bound = max(
            0,
            self.pattern_start + self.ace_note_list[-1].pos + self.ace_note_list[-1].dur + 120,
        )
        segments = []
        for seg in curve.split_into_segments(-100):
            if seg[-1].x < self.first_bar_ticks:
                continue
            start_ticks = seg[0].x - self.first_bar_ticks if seg[0].x > self.first_bar_ticks else 0
            end_ticks = seg[-1].x - self.first_bar_ticks
            if start_ticks <= right_bound and end_ticks >= left_bound:
                segments.append(seg)
        for segment in segments:
            start_point = (
                binary_find_last(
                    segment,
                    lambda point: (0 <= point.x - self.first_bar_ticks <= left_bound),
                )
                or segment[0]
            )
            end_point = (
                binary_find_first(
                    segment,
                    lambda point: (right_bound <= point.x - self.first_bar_ticks),
                )
                or segment[-1]
            )
            ace_curve = AcepParamCurve(
                offset=round(start_point.x - self.first_bar_ticks - self.pattern_start)
            )
            curve_end = round(end_point.x - self.first_bar_ticks - self.pattern_start)
            tick_step = (end_point.x - start_point.x) / (curve_end - ace_curve.offset)
            tick = float(start_point.x)
            while tick < self.first_bar_ticks:
                ace_curve.offset += 1
                tick += tick_step
            tick = max(self.first_bar_ticks, tick)
            tick_step = (end_point.x - tick) / (curve_end - ace_curve.offset)
            while tick < left_bound:
                ace_curve.offset += 1
                tick += tick_step
            pos = ace_curve.offset
            while pos <= min(right_bound, curve_end):
                ace_curve.values.append(mapping_func(self.get_value_from_segment(segment, tick)))
                pos += 1
                tick += tick_step
            ace_curves.root.append(ace_curve)
        return ace_curves

    @staticmethod
    def get_value_from_segment(segment: list[Point], ticks: float) -> float:
        if (left_point := binary_find_last(segment, lambda point: point.x <= ticks)) is None:
            return segment[0].y
        if (right_point := binary_find_first(segment, lambda point: point.x > ticks)) is None:
            return segment[-1].y
        ratio = (ticks - left_point.x) / (right_point.x - left_point.x)
        return (1 - ratio) * left_point.y + ratio * right_point.y



from dataclasses import dataclass
from math import ceil

from libresvip.core.exceptions import NoTrackError
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Project, SingingTrack
from libresvip.utils.translation import gettext_lazy as _

from .options import OutputOptions
from .utils import pinyin_util
from .utils.gender_param_utils import GenderParamUtils
from .utils.models.ds_file import DsItem
from .utils.models.ds_project import DsProjectModel
from .utils.note_list_util import encode_notes
from .utils.pitch_param_utils import PitchParamUtils


@dataclass
class DiffSingerGenerator:
    options: OutputOptions
    trailing_space: float = 0.05

    def generate(self, project: Project) -> DsItem:
        pinyin_util.load_phoneme_table(self.options.dict_name)
        synchronizer = TimeSynchronizer(project.song_tempo_list)
        if self.options.track_index < 0:
            singing_track = next(
                (track for track in project.track_list if isinstance(track, SingingTrack)),
                None,
            )
        else:
            singing_track = project.track_list[self.options.track_index]
        if singing_track is None:
            msg = _("No singing track found")
            raise NoTrackError(msg)
        os_notes = singing_track.note_list
        ds_project = DsProjectModel()
        ds_project.note_list = encode_notes(os_notes, synchronizer, self.trailing_space)
        total_duration = ceil(sum(note.duration for note in ds_project.note_list) * 1000)
        os_pitch_param_curve = singing_track.edited_params.pitch
        ds_project.pitch_param_curve = PitchParamUtils.encode(os_pitch_param_curve, total_duration)

        if self.options.export_gender:
            os_gender_param_curve = singing_track.edited_params.gender
            ds_project.gender_param_curve = GenderParamUtils.encode(
                os_gender_param_curve, total_duration
            )

        return ds_project.to_param_model()



import dataclasses

from libresvip.core.constants import DEFAULT_PHONEME, TICKS_IN_BEAT
from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.music_math import note2midi

from .models.enums import StartStop, StartStopContinue
from .models.mxml4 import ScorePart, ScorePartwise
from .options import InputOptions


@dataclasses.dataclass
class MusicXMLParser:
    options: InputOptions

    def parse_project(self, mxml: ScorePartwise) -> Project:
        part_nodes = mxml.part

        master_track = next((part for part in part_nodes if part.measure), None)
        assert master_track is not None
        (
            time_signatures,
            tempos,
            import_tick_rate,
        ) = self.parse_master_track(master_track)
        time_signatures = time_signatures or [TimeSignature()]
        tempos = tempos or [SongTempo()]

        tracks = []
        for index, part in enumerate(part_nodes):
            score_part = None
            if mxml.part_list is not None and part.id is not None:
                score_part = next(
                    (item for item in mxml.part_list.score_part if item.id == part.id),
                    None,
                )
            tracks.append(self.parse_track(index, part, score_part, import_tick_rate))

        return Project(
            track_list=tracks,
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
        )

    def parse_master_track(
        self, part_node: ScorePartwise.Part
    ) -> tuple[list[TimeSignature], list[SongTempo], float]:
        measure_nodes = part_node.measure
        assert measure_nodes[0].attributes[0].divisions is not None
        divisions = int(measure_nodes[0].attributes[0].divisions)
        import_tick_rate = TICKS_IN_BEAT / divisions
        tempos = []
        time_signatures = []
        tick_position = 0
        current_time_signature = TimeSignature()
        for i, measure_node in enumerate(measure_nodes):
            if (
                len(measure_node.attributes)
                and len(measure_node.attributes[0].time)
                and (time_signature_node := measure_node.attributes[0].time[0])
            ):
                bar_index = int(time_signature_node.number or i)
                numerator = int(time_signature_node.beats[0])
                denominator = int(time_signature_node.beat_type[0])
                current_time_signature = TimeSignature(
                    bar_index=bar_index, numerator=numerator, denominator=denominator
                )
                time_signatures.append(current_time_signature)

            sound_nodes = measure_node.sound
            if len(sound_nodes) and (sound_node := sound_nodes[0]).tempo:
                tempo = SongTempo(
                    position=tick_position,
                    bpm=float(sound_node.tempo),
                )
                tempos.append(tempo)

            tick_position += round(current_time_signature.bar_length())
        return time_signatures, tempos, import_tick_rate

    def parse_track(
        self,
        track_index: int,
        part_node: ScorePartwise.Part,
        score_part: ScorePart | None,
        import_tick_rate: float,
    ) -> SingingTrack:
        track_name = (
            score_part.part_name.value
            if score_part is not None and score_part.part_name is not None
            else f"Track {track_index + 1}"
        )
        notes = []
        is_inside_note = False
        measure_nodes = part_node.measure
        tick_position = 0
        for measure_node in measure_nodes:
            note_nodes = measure_node.note
            for note_node in note_nodes:
                duration_nodes = note_node.duration
                duration = (
                    int(float(duration_nodes[0]) * import_tick_rate)
                    if len(duration_nodes)
                    else None
                )
                if not duration:
                    if note_node.grace:
                        continue
                    else:
                        msg = "Duration not found"
                        raise ValueError(msg)

                rest_nodes = note_node.rest
                if rest_nodes:
                    tick_position += duration
                    continue

                pitch_node = note_node.pitch[0]
                step = pitch_node.step
                assert step is not None
                assert pitch_node.octave is not None
                alter_node = pitch_node.alter
                alter = int(alter_node) if alter_node else 0
                octave = int(pitch_node.octave)
                key = note2midi(f"{step.value}{octave}") + alter

                lyric_nodes = note_node.lyric
                if any(
                    slur.type_value in [StartStopContinue.CONTINUE, StartStopContinue.STOP]
                    for notation in note_node.notations
                    for slur in notation.slur
                ):
                    lyric = "-"
                elif len(lyric_nodes) and len(lyric_nodes[0].text):
                    lyric = lyric_nodes[0].text[0].value
                else:
                    lyric = DEFAULT_PHONEME

                if not is_inside_note:
                    note = Note(
                        key_number=key,
                        lyric=lyric,
                        start_pos=tick_position,
                        length=duration,
                    )
                    notes.append(note)
                else:
                    notes[-1] = notes[-1].model_copy(
                        update={
                            "length": notes[-1].length + duration,
                        }
                    )

                tick_position += duration

                tie_nodes = note_node.tie
                if len(tie_nodes) and (tie_node := tie_nodes[0]) and tie_node.type_value:
                    if tie_node.type_value == StartStop.START:
                        is_inside_note = True
                    elif tie_node.type_value == StartStop.STOP:
                        is_inside_note = False

        return SingingTrack(
            title=track_name,
            note_list=notes,
        )



import typer

from .commands import conf_app, plugin_app, proj_app

app = typer.Typer()

app.add_typer(conf_app, name="conf")
app.add_typer(plugin_app, name="plugin")
app.add_typer(proj_app, name="proj")



import math

from libresvip.model.point import Point


def perpendicular_distance(pt: Point, line_start: Point, line_end: Point) -> float:
    dx: float = line_end.x - line_start.x
    dy: float = line_end.y - line_start.y

    # Normalize
    mag = math.hypot(dx, dy)
    if mag > 0.0:
        dx /= mag
        dy /= mag
    pvx = pt.x - line_start.x
    pvy = pt.y - line_start.y

    # Get dot product (project pv onto normalized direction)
    pvdot = dx * pvx + dy * pvy

    # Scale line direction vector and subtract it from pv
    ax = pvx - pvdot * dx
    ay = pvy - pvdot * dy

    return math.hypot(ax, ay)


def simplify_shape(point_list: list[Point], epsilon: float) -> list[Point]:
    if len(point_list) < 2:
        return point_list

    # Find the point with the maximum distance from line between start and end
    dmax = 0.0
    index = 0
    end = len(point_list) - 1
    for i in range(1, end):
        d = perpendicular_distance(point_list[i], point_list[0], point_list[end])
        if d > dmax:
            index = i
            dmax = d

    if dmax <= epsilon:
        # Just return start and end points
        return [point_list[0], point_list[-1]]
    first_line = point_list[: index + 1]
    last_line = point_list[index:]
    rec_results1 = simplify_shape(first_line, epsilon)
    rec_results2 = simplify_shape(last_line, epsilon)

    # build the result list
    return rec_results1[:-1] + rec_results2


def simplify_shape_to(point_list: list[Point], max_point_count: int) -> list[Point]:
    step = 0.05
    epsilon = step
    while True:
        result = simplify_shape(point_list, epsilon)
        if len(result) < max_point_count:
            return result
        epsilon += step



import dataclasses
import itertools
import operator
from collections.abc import MutableSequence
from urllib.parse import urljoin

from google.protobuf import any_pb2

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.tick_counter import shift_tempo_list
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import db_to_float, ratio_to_db

from .constants import TYPE_URL_BASE, Svip3TrackType
from .model import (
    Svip3AudioTrack,
    Svip3Note,
    Svip3Project,
    Svip3SingingPattern,
    Svip3SingingTrack,
    Svip3SongBeat,
    Svip3SongTempo,
)
from .options import InputOptions
from .singers import singers_data


@dataclasses.dataclass
class Svip3Parser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)
    song_tempo_list: list[SongTempo] = dataclasses.field(init=False)

    def parse_project(self, svip3_project: Svip3Project) -> Project:
        time_signature_list = self.parse_time_signatures(svip3_project.beat_list)
        self.first_bar_length = round(time_signature_list[0].bar_length())
        self.song_tempo_list = self.parse_song_tempos(svip3_project.tempo_list)
        tracks = self.parse_tracks(svip3_project.track_list)
        return Project(
            time_signature_list=time_signature_list,
            song_tempo_list=self.song_tempo_list,
            track_list=tracks,
        )

    @staticmethod
    def parse_time_signatures(
        beat_list: MutableSequence[Svip3SongBeat],
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=beat.pos,
                numerator=beat.beat_size.numerator,
                denominator=beat.beat_size.denominator,
            )
            for beat in beat_list
        ]

    def parse_song_tempos(self, tempo_list: MutableSequence[Svip3SongTempo]) -> list[SongTempo]:
        return shift_tempo_list(
            [
                SongTempo(
                    position=tempo.pos,
                    bpm=tempo.tempo / 100.0,
                )
                for tempo in tempo_list
            ],
            self.first_bar_length,
        )

    def parse_tracks(self, track_list: MutableSequence[any_pb2.Any]) -> list[Track]:
        tracks = []
        for track in track_list:
            if track.type_url == urljoin(TYPE_URL_BASE, Svip3TrackType.SINGING_TRACK):
                singing_track = Svip3SingingTrack.deserialize(track.value)
                tracks.append(self.parse_singing_track(singing_track))
            elif self.options.import_instrumental_track and track.type_url == urljoin(
                TYPE_URL_BASE, Svip3TrackType.AUDIO_TRACK
            ):
                audio_track = Svip3AudioTrack.deserialize(track.value)
                if xstudio_audio_track := self.parse_audio_track(audio_track):
                    tracks.append(xstudio_audio_track)
        return tracks

    def parse_audio_track(self, audio_track: Svip3AudioTrack) -> InstrumentalTrack | None:
        audio_file_path = None
        offset = 0
        if len(audio_track.pattern_list):
            first_pattern = audio_track.pattern_list[0]
            audio_file_path = first_pattern.audio_file_path
            offset = first_pattern.real_pos
        if audio_file_path is not None:
            return InstrumentalTrack(
                audio_file_path=audio_file_path,
                offset=offset,
                pan=self.parse_pan(audio_track.pan),
                title=audio_track.name,
                mute=audio_track.mute or False,
                solo=audio_track.solo or False,
                volume=self.to_linear_volume(audio_track.volume),
            )

    @staticmethod
    def parse_pan(pan: float) -> float:
        return (pan or 0) / 10.0

    @staticmethod
    def to_linear_volume(gain: float) -> float:
        if gain >= 0:
            return min(
                2.0,
                gain / ratio_to_db(4.0) + 1.0,
            )
        else:
            return db_to_float(gain)

    def parse_singing_track(self, singing_track: Svip3SingingTrack) -> SingingTrack:
        pattern_list = sorted(singing_track.pattern_list, key=operator.attrgetter("pos"))
        return SingingTrack(
            title=singing_track.name,
            mute=singing_track.mute,
            solo=singing_track.solo,
            volume=self.to_linear_volume(singing_track.volume),
            pan=self.parse_pan(singing_track.pan),
            ai_singer_name=singers_data.get(singing_track.ai_singer_id, ""),
            note_list=self.parse_notes(pattern_list),
            edited_params=self.parse_edited_params(pattern_list),
        )

    def parse_notes(self, pattern_list: list[Svip3SingingPattern]) -> list[Note]:
        note_list: list[Note] = []
        for pattern in pattern_list:
            offset = pattern.real_pos
            left = pattern.play_pos + offset
            right = left + pattern.play_dur
            visible_notes = [
                note
                for note in pattern.note_list
                if left <= note.start_pos + offset <= right - note.width_pos
            ]
            note_list.extend(self.parse_note(note, offset) for note in visible_notes)
        return note_list

    def parse_edited_params(self, pattern_list: list[Svip3SingingPattern]) -> Params:
        params = Params()
        if self.options.import_pitch:
            params.pitch = self.parse_pitch_curve(pattern_list)
        return params

    def parse_note(self, svip3_note: Svip3Note, offset: int) -> Note:
        return Note(
            start_pos=svip3_note.start_pos + offset,
            length=svip3_note.width_pos,
            key_number=svip3_note.key_index,
            lyric=svip3_note.lyric,
            pronunciation=self.parse_pronunciation(svip3_note),
            head_tag=self.parse_head_tag(svip3_note),
            edited_phones=self.parse_edited_phones(svip3_note),
        )

    @staticmethod
    def parse_pronunciation(svip3_note: Svip3Note) -> str:
        if not svip3_note.pronouncing and "-" not in svip3_note.lyric:
            return svip3_note.lyric
        return svip3_note.pronouncing

    @staticmethod
    def parse_head_tag(svip3_note: Svip3Note) -> str | None:
        if svip3_note.sil_len > 0:
            return "0"
        elif svip3_note.sp_len > 0:
            return "V"
        return None

    def parse_edited_phones(self, svip3_note: Svip3Note) -> Phones | None:
        if svip3_note.consonant_len > 0:
            return Phones(
                head_length_in_secs=(
                    svip3_note.consonant_len / TICKS_IN_BEAT * 60.0 / self.song_tempo_list[0].bpm
                )
            )
        return None

    def parse_pitch_curve(self, pattern_list: list[Svip3SingingPattern]) -> ParamCurve:
        curves = [self.parse_pattern_curve(pattern) for pattern in pattern_list]
        curve = ParamCurve()
        curve.points.append(Point.start_point())
        curve.points.root.extend(self.merge_param_curves(curves))
        curve.points.append(Point.end_point())
        return curve

    @staticmethod
    def merge_param_curves(curves: list[ParamCurve]) -> list[Point]:
        return [*itertools.chain.from_iterable(curve.points.root for curve in curves)]

    @staticmethod
    def get_visible_range(pattern: Svip3SingingPattern) -> tuple[int, int]:
        offset = pattern.real_pos
        left = pattern.play_pos + offset
        right = left + pattern.play_dur
        return left, right

    def parse_pattern_curve(self, pattern: Svip3SingingPattern) -> ParamCurve:
        curve = ParamCurve()
        left, right = self.get_visible_range(pattern)
        visible_nodes = [node for node in pattern.edited_pitch_line if left <= node.pos <= right]
        curve.points.root.append(Point(left, -100))
        for node in visible_nodes:
            pos = node.pos + self.first_bar_length
            val = round(node.value * 100 - 50) if node.value != -1.0 else -100
            curve.points.append(Point(pos, val))
        curve.points.root.append(Point(right, -100))
        return curve



from dataclasses import InitVar, dataclass, field
from typing import Literal


@dataclass
class DsPhonemeItem:
    phoneme: str = ""
    duration: float | None = None
    note_name: str = ""


@dataclass
class DsPhoneme:
    consonant: DsPhonemeItem | None = None
    vowel: DsPhonemeItem = field(default_factory=DsPhonemeItem)


@dataclass
class DsNote:
    duration: float
    ds_phoneme: DsPhoneme
    lyric: str = ""
    note_name: str = ""

    @property
    def is_slur(self) -> bool:
        return "-" in self.lyric


@dataclass
class AspirationDsPhoneme(DsPhoneme):
    _duration: InitVar[float] = 0.0

    def __post_init__(self, _duration: float) -> None:
        self.vowel = DsPhonemeItem("AP", _duration, "rest")


@dataclass
class AspirationDsNote(DsNote):  # type: ignore[override]
    ds_phoneme: AspirationDsPhoneme
    lyric: Literal["AP"] = "AP"
    note_name: Literal["rest"] = "rest"


@dataclass
class RestDsPhoneme(DsPhoneme):
    _duration: InitVar[float] = 0.0

    def __post_init__(self, _duration: float) -> None:
        self.vowel = DsPhonemeItem("SP", _duration, "rest")


@dataclass
class RestDsNote(DsNote):  # type: ignore[override]
    ds_phoneme: RestDsPhoneme
    lyric: Literal["SP"] = "SP"
    note_name: Literal["rest"] = "rest"



import asyncio
import fnmatch
import platform
import time
from collections.abc import Awaitable, Sequence
from functools import partial
from importlib.resources import as_file
from typing import TYPE_CHECKING

import httpx
from desktop_notifier import Button, DesktopNotifier, Notification
from loguru import logger
from packaging.version import Version
from PySide6.QtCore import QObject, QTimer, Slot
from PySide6.QtQml import QmlElement

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

import libresvip
from libresvip.core.constants import PACKAGE_NAME, app_dir, res_dir
from libresvip.utils.translation import gettext_lazy as _

from .application import app, event_loop
from .url_opener import open_path, open_url

if TYPE_CHECKING:
    from contextlib import AbstractContextManager
    from pathlib import Path

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


@QmlElement
class Notifier(QObject):
    def __init__(self) -> None:
        super().__init__()
        self.request_timeout = 30
        self.notify_timeout = 5
        self.last_notify_time: float | None = None
        try:
            if platform.system() == "Darwin":
                from desktop_notifier.backends.macos_support import (
                    is_bundle,
                    macos_version,
                )

                if macos_version >= Version("10.14") and not is_bundle():
                    msg = "macOS app is not signed"
                    raise RuntimeError(msg)  # noqa: TRY301

            icon_path: AbstractContextManager[Path] = as_file(res_dir / "libresvip.ico")
            app.aboutToQuit.connect(lambda: icon_path.__exit__(None, None, None))
            self.notifier = DesktopNotifier(app_name=PACKAGE_NAME, app_icon=icon_path.__enter__())
        except Exception:
            self.notifier = None

    async def download_release(self, url: str, filename: str) -> None:
        app_dir.user_downloads_path.mkdir(parents=True, exist_ok=True)
        async with httpx.AsyncClient(follow_redirects=True, verify=False) as client:
            try:
                await self.clear_all_messages_async()
                await self.notify_async(
                    title=_("Downloading"),
                    message=_("Please wait..."),
                    send_timeout=self.request_timeout,
                )
                with (app_dir.user_downloads_path / filename).open("wb") as f:
                    async with client.stream("GET", url) as response:
                        response.raise_for_status()
                        async for chunk in response.aiter_bytes():
                            f.write(chunk)
                            f.flush()
                open_path(app_dir.user_downloads_path)
            except httpx.HTTPError:
                await self.notify_async(
                    title=_("Error occurred while Downloading"),
                    message=_("Failed to download file {}. Please try again later.").format(
                        filename
                    ),
                    send_timeout=self.notify_timeout,
                )

    async def _check_for_updates(self) -> None:
        failed = False
        logger.info("Checking for updates...")
        async with httpx.AsyncClient(
            follow_redirects=True, timeout=self.request_timeout, verify=False
        ) as client:
            try:
                await self.clear_all_messages_async()
                await self.notify_async(
                    title=_("Checking for Updates"),
                    message=_("Please wait..."),
                    send_timeout=self.notify_timeout,
                )
                resp = await client.get(
                    "https://api.github.com/repos/SoulMelody/LibreSVIP/releases/latest"
                )
                if resp.status_code == 200:
                    data = resp.json()
                    local_version = Version(libresvip.__version__)
                    remote_version = Version(data["tag_name"].removeprefix("v"))
                    if remote_version > local_version:
                        uname = platform.uname()
                        arch = uname.machine.lower()
                        buttons = [
                            Button(
                                _("Open in Browser"),
                                partial(
                                    open_url,
                                    data["html_url"],
                                ),
                            )
                        ]
                        if arch.endswith("64"):
                            asset = None
                            if "aarch" in arch and uname.system != "Linux":
                                pass
                            elif uname.system == "Windows":
                                python_compiler = platform.python_compiler()
                                if python_compiler.startswith("GCC") and "arm" not in arch:
                                    asset = next(
                                        (
                                            asset
                                            for asset in data["assets"]
                                            if fnmatch.fnmatch(
                                                asset["name"],
                                                "LibreSVIP-*.msys2-*.7z",
                                            )
                                        ),
                                        None,
                                    )
                                else:
                                    asset = next(
                                        (
                                            asset
                                            for asset in data["assets"]
                                            if fnmatch.fnmatch(
                                                asset["name"],
                                                f"LibreSVIP-*.win-{arch}.*",
                                            )
                                        ),
                                        None,
                                    )
                            elif uname.system == "Linux":
                                asset = next(
                                    (
                                        asset
                                        for asset in data["assets"]
                                        if fnmatch.fnmatch(
                                            asset["name"],
                                            f"LibreSVIP-*.linux-{arch}.tar.gz",
                                        )
                                    ),
                                    None,
                                )
                            elif uname.system == "Darwin":
                                asset = next(
                                    (
                                        asset
                                        for asset in data["assets"]
                                        if fnmatch.fnmatch(
                                            asset["name"],
                                            f"LibreSVIP-*.macos-{arch}.dmg",
                                        )
                                    ),
                                    None,
                                )
                            if asset:
                                buttons.append(
                                    Button(
                                        _("Download"),
                                        lambda: asyncio.create_task(
                                            self.download_release(
                                                asset["browser_download_url"],
                                                asset["name"],
                                            ),
                                        ),
                                    ),
                                )
                        await self.notify_async(
                            title=_("Update Available"),
                            message=_("New version {} is available.").format(remote_version),
                            buttons=buttons,
                        )
                    else:
                        await self.notify_async(
                            title=_("No Updates"),
                            message=_("You are using the latest version {}.").format(local_version),
                        )
                else:
                    failed = True
            except httpx.HTTPError:
                failed = True
            if failed:
                await self.notify_async(
                    title=_("Error occurred while Checking for Updates"),
                    message=_("Failed to check for updates. Please try again later."),
                    send_timeout=self.notify_timeout,
                )

    @Slot(result=None)
    def check_for_updates(self) -> None:
        timer = QTimer(self)
        timer.interval = 100

        def ensure_running_loop() -> None:
            if event_loop.is_running():
                event_loop.create_task(self._check_for_updates())
                timer.stop()

        timer.timeout.connect(ensure_running_loop)
        timer.start()

    async def notify_async(
        self,
        title: str,
        message: str,
        buttons: Sequence[Button] = (),
        send_timeout: int = -1,
    ) -> Awaitable[Notification] | None:
        try:
            if self.last_notify_time is None:
                pass
            elif (elapsed := time.time() - self.last_notify_time) < 1:
                await asyncio.sleep(1 - elapsed)
            self.last_notify_time = time.time()
            return await self.notifier.send(
                title=title,
                message=message,
                buttons=buttons,
                timeout=send_timeout,
            )
        except Exception as e:
            logger.exception(e)

    async def clear_all_messages_async(self) -> None:
        try:
            if len(await self.notifier.get_current_notifications()):
                await self.notifier.clear_all()
        except Exception as e:
            logger.exception(e)

    @Slot(str)
    def open_link(self, url: str) -> None:
        open_url(url)



import dataclasses
import functools
import math
import operator
from typing import cast

import more_itertools
import portion

from libresvip.core.tick_counter import shift_tempo_list
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import HermiteInterpolator, db_to_float, ratio_to_db
from libresvip.utils.translation import gettext_lazy as _

from .model import (
    TuneLabAudioPart,
    TuneLabMidiPart,
    TuneLabNote,
    TuneLabPoints,
    TuneLabProject,
    TuneLabTempo,
    TuneLabTimeSignature,
    TuneLabTrack,
    TuneLabVibrato,
)
from .options import InputOptions


@dataclasses.dataclass
class TuneLabParser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def parse_project(self, project: TuneLabProject) -> Project:
        time_signature_list = self.parse_time_signatures(project.time_signatures)
        self.first_bar_length = round(time_signature_list[0].bar_length())
        tempo_list = self.parse_tempos(project.tempos)
        self.synchronizer = TimeSynchronizer(tempo_list)
        return Project(
            song_tempo_list=tempo_list,
            time_signature_list=time_signature_list,
            track_list=self.parse_tracks(project.tracks),
        )

    def parse_time_signatures(
        self, time_signatures: list[TuneLabTimeSignature]
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=time_signature.bar_index,
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            for time_signature in time_signatures
        ]

    def parse_tempos(self, tempos: list[TuneLabTempo]) -> list[SongTempo]:
        return shift_tempo_list(
            [SongTempo(position=int(tempo.pos), bpm=tempo.bpm) for tempo in tempos],
            self.first_bar_length,
        )

    def parse_tracks(self, tracks: list[TuneLabTrack]) -> list[Track]:
        track_list = []
        for track in tracks:
            for part in track.parts:
                if isinstance(part, TuneLabAudioPart) and self.options.import_instrumental_track:
                    track_list.append(
                        InstrumentalTrack(
                            audio_file_path=part.path,
                            title=part.name,
                            offset=int(part.pos),
                            volume=self.parse_volume(track.gain),
                            pan=track.pan,
                            mute=track.mute,
                            solo=track.solo,
                        )
                    )
                elif isinstance(part, TuneLabMidiPart) and len(part.notes):
                    if (
                        track_list
                        and isinstance(track_list[-1], SingingTrack)
                        and (
                            not track_list[-1].note_list
                            or track_list[-1].note_list[-1].end_pos <= int(part.pos)
                        )
                    ):
                        singing_track = track_list[-1]
                    else:
                        singing_track = SingingTrack(
                            title=part.name,
                            volume=self.parse_volume(track.gain),
                            pan=track.pan,
                            mute=track.mute,
                            solo=track.solo,
                        )
                        track_list.append(singing_track)
                    singing_track.note_list.extend(self.parse_notes(part.notes, int(part.pos)))
                    if self.options.import_pitch:
                        (
                            vibrato_base_interval_dict,
                            vibrato_envelope_interval_dict,
                        ) = self.parse_vibrato(part)
                        if pitch_points := self.parse_pitch(
                            part.pitch,
                            int(part.pos),
                            vibrato_base_interval_dict,
                            vibrato_envelope_interval_dict,
                        ):
                            singing_track.edited_params.pitch.points.root.extend(pitch_points)
        return track_list

    @staticmethod
    def parse_volume(gain: float) -> float:
        if gain >= 0:
            return min(gain / (ratio_to_db(4)) + 1.0, 2.0)
        else:
            return db_to_float(gain)

    @staticmethod
    def parse_notes(notes: list[TuneLabNote], offset: int) -> list[Note]:
        note_list: list[Note] = []
        if len(notes):
            next_pos = None
            for tlp_note in notes[::-1]:
                normalized_duration = int(tlp_note.dur)
                if next_pos is not None:
                    distance = next_pos - tlp_note.pos
                    if distance < normalized_duration:
                        normalized_duration = int(distance)
                        if normalized_duration > 0:
                            show_warning(_("Note overlap detected, cutting note ") + tlp_note.lyric)
                if normalized_duration > 0:
                    note_list.insert(
                        0,
                        Note(
                            start_pos=int(tlp_note.pos + offset),
                            length=normalized_duration,
                            key_number=tlp_note.pitch,
                            lyric=tlp_note.lyric,
                            pronunciation=tlp_note.pronunciation,
                        ),
                    )
                else:
                    show_warning(_("Note overlap detected, skipping note ") + tlp_note.lyric)
                next_pos = int(tlp_note.pos)
        return note_list

    @staticmethod
    def vibrato_value(seconds: float, vibrato_start: float, vibrato: TuneLabVibrato) -> float:
        return (
            math.sin(math.pi * (2 * (seconds - vibrato_start) * vibrato.frequency - vibrato.phase))
            * vibrato.amplitude
        )

    def parse_vibrato(
        self, part: TuneLabMidiPart
    ) -> tuple[PiecewiseIntervalDict, PiecewiseIntervalDict]:
        vibrato_base_interval_dict = PiecewiseIntervalDict()
        vibrato_envelope_interval_dict = PiecewiseIntervalDict()
        for vibrato in part.vibratos:
            vibrato_start = self.synchronizer.get_actual_secs_from_ticks(int(vibrato.pos))
            vibrato_end = self.synchronizer.get_actual_secs_from_ticks(
                int(vibrato.pos + vibrato.dur)
            )
            vibrato_base_interval_dict[portion.closed(vibrato_start, vibrato_end)] = (
                functools.partial(
                    self.vibrato_value,
                    vibrato_start=vibrato_start,
                    vibrato=vibrato,
                )
            )
        if part.automations.get("VibratoEnvelope"):
            for value, ticks_group in more_itertools.groupby_transform(
                part.automations["VibratoEnvelope"].values.root,
                keyfunc=operator.attrgetter("value"),
                valuefunc=operator.attrgetter("pos"),
            ):
                ticks_group = list(ticks_group)
                group_start = self.synchronizer.get_actual_secs_from_ticks(int(ticks_group[0]))
                group_end = self.synchronizer.get_actual_secs_from_ticks(int(ticks_group[-1] + 5))
                vibrato_envelope_interval_dict[
                    portion.closedopen(
                        group_start,
                        group_end,
                    )
                ] = max(value + 1, 0)
        return vibrato_base_interval_dict, vibrato_envelope_interval_dict

    def parse_pitch(
        self,
        pitch: list[TuneLabPoints],
        offset: int,
        vibrato_base_interval_dict: PiecewiseIntervalDict,
        vibrato_envelope_interval_dict: PiecewiseIntervalDict,
    ) -> list[Point]:
        points: list[Point] = [Point.start_point()]
        for pitch_part in pitch:
            anchor_group = [point for point in pitch_part.root if not math.isnan(point.value)]
            if len(anchor_group) < 2:
                continue
            interpolator = HermiteInterpolator(
                points=cast("list[tuple[float, float]]", anchor_group)
            )
            xs = list(
                more_itertools.numeric_range(anchor_group[0].pos, anchor_group[-1].pos + 1, 5)
            )
            ys = interpolator.interpolate(xs)
            for i, (x, y) in enumerate(zip(xs, ys)):
                pitch_pos = int(x) + offset
                if i == 0:
                    points.append(
                        Point(
                            x=pitch_pos + self.first_bar_length,
                            y=-100,
                        )
                    )
                pitch_secs = self.synchronizer.get_actual_secs_from_ticks(pitch_pos)
                pitch_value = y
                if (vibrato_value := vibrato_base_interval_dict.get(pitch_secs)) is not None:
                    vibrato_value *= vibrato_envelope_interval_dict.get(pitch_secs, 1)
                    pitch_value += vibrato_value
                points.append(
                    Point(
                        x=pitch_pos + self.first_bar_length,
                        y=round(pitch_value * 100),
                    )
                )
            points.append(
                Point(
                    x=points[-1].x,
                    y=-100,
                )
            )
        points.append(Point.end_point())
        return points



from construct import (
    Bytes,
    BytesInteger,
    Const,
    FixedSized,
    Float64l,
    GreedyBytes,
    If,
    Int16sl,
    Int16ul,
    Struct,
    this,
)
from construct import Enum as CSEnum

Int32sl = BytesInteger(4, swapped=True, signed=True)
Int32ul = BytesInteger(4, swapped=True, signed=False)

VocalShifterLabel = Struct(
    "start_tick" / Int32ul,
    "end_tick" / Int32ul,
    "index" / Int32ul,
    "padding" / Bytes(20),
    "name" / FixedSized(64, GreedyBytes),
)


VocalShifterLabels = Struct(
    "magic" / Const(b"Labl"),
    "size" / Int32ul,
    "labels" / VocalShifterLabel[this._.header.label_count],
)


VocalShifterNote = Struct(
    "start_tick" / Int32ul,
    "length" / Int32ul,
    "pitch" / Int32ul,
    "padding" / Bytes(20),
)


VocalShifterNotes = Struct(
    "magic" / Const(b"Note"),
    "size" / Int32ul,
    "notes" / VocalShifterNote[this._.header.notes_count],
)


VocalShifterTime = Struct(
    "magic" / Const(b"Time"),
    "size" / Int32ul,
    "original_time" / Float64l,
    "time" / Float64l,
)

VocalShifterControlPoint = Struct(
    "magic" / Const(b"Ctrp"),
    "size" / Int32ul,
    "pit_array" / Int16ul[4],
    "padding1" / Bytes(12),
    "ori_pit" / Int16ul,
    "pit" / Int16ul,
    "frm" / Int16sl,
    "bre" / Int16sl,
    "eq1" / Int16sl,
    "eq2" / Int16sl,
    "ori_dyn" / Float64l,
    "dyn" / Float64l,
    "vol" / Float64l,
    "pan" / Float64l,
    "padding2" / Bytes(18),
    "heq" / Int16sl,
    "padding3" / Bytes(12),
)

VocalShifterPatternType = CSEnum(
    Int32ul,
    WAVE=0,
    MIDI=1,
)


VocalShifterAnalysisMethod = CSEnum(
    Int16ul,
    Monophonic=0,
    MonophonicVocal=1,
    Polyphonic=2,
    Rhythm=4,
    World=8,
)


VocalShifterTemperament = CSEnum(
    Int32ul,
    EqualTemperament=0,
    JustIntonation=1,
    PythagoreanTuning=2,
    MeantoneTemperament=3,
)


VocalShifterPatternHeader = Struct(
    "magic" / Const(b"Itmp"),
    "size" / Int32ul,
    "sample_count" / Int32ul,
    "sample_rate" / Int32ul,
    "channels" / Int32ul,
    "pattern_type" / VocalShifterPatternType,
    "points_per_second" / Int32ul,
    "points_count" / Int32ul,
    "padding1" / Bytes(8),
    "key_min" / Int16ul,
    "key_scale" / Int16ul,  # key_max = key_min + key_scale
    "padding2" / Bytes(12),
    "analysis_method" / VocalShifterAnalysisMethod,
    "padding3" / Bytes(78),
    "spectrum_key_min" / Int16ul,
    "spectrum_key_scale" / Int16ul,
    "spectrum_points_density" / Int16ul,
    "spectrum_unknown1" / Int16ul,
    "spectrum_points_count" / Int32ul,
    "spectrum_unknown2" / Int32ul,
    "notes_count" / Int32ul,
    "label_count" / Int32ul,
    "padding4" / Bytes(8),
    "eq1" / Int16sl[16],
    "eq2" / Int16sl[16],
    "heq" / Int16sl[16],
)


VocalShifterSpectrumData = Struct(
    "magic" / Const(b"Spct"),
    "size" / Int32ul,
    "data"
    / Int16ul[
        this._.header.spectrum_points_count
        * this._.header.spectrum_points_density
        * this._.header.spectrum_key_scale
    ],
)


VocalShifterPatternData = Struct(
    "magic" / Const(b"ITMD"),
    "size" / Int32ul,
    "header" / VocalShifterPatternHeader,
    "points" / VocalShifterControlPoint[this.header.points_count],
    "start_time" / VocalShifterTime,
    "end_time" / VocalShifterTime,
    "spectrum"
    / If(
        this.header.spectrum_points_count > 0,
        VocalShifterSpectrumData,
    ),
    "notes" / VocalShifterNotes,
    "labels" / If(this.header.label_count > 0, VocalShifterLabels),
)

VocalShifterPatternMetadata = Struct(
    "magic" / Const(b"ITMP"),
    "size" / Int32ul,
    "path_and_ext" / Bytes(256),
    "pattern_index" / Int32ul,
    "track_index" / Int32ul,
    "padding1" / Bytes(4),
    "offset_correction" / Int32sl,
    "offset_samples" / Float64l,
    "measure_offset" / Float64l,
    "base_freq" / Float64l,
    "key" / Int32ul,
    "temperament" / VocalShifterTemperament,
    "padding2" / Bytes(16),
    "modified_base_freq" / Float64l,
    "is_sharp" / Int32ul,
    "mrp_auto_set" / Int32ul,
    "reserved" / Bytes(176),
)

VocalShifterTrackColor = CSEnum(
    Int32ul,
    Red=0,
    Orange=1,
    Yellow=2,
    Green=3,
    LightBlue=4,
    Blue=5,
    Purple=6,
    Pink=7,
)

VocalShifterTrackMetadata = Struct(
    "magic" / Const(b"TRKP"),
    "size" / Int32ul,
    "name" / FixedSized(64, GreedyBytes),
    "volume" / Float64l,
    "pan" / Float64l,
    "mute" / Int32ul,
    "solo" / Int32ul,
    "phase_pre" / Int32ul,
    "color" / VocalShifterTrackColor,
    "phase_next" / Int32ul,
    "morph_group" / Int32ul,
    "mute_flag" / Int32ul,
    "reserved" / Bytes(148),
)


VocalShifterProjectMetadata = Struct(
    "magic" / Const(b"PRJP"),
    "size" / Int32ul,
    "track_count" / Int32ul,
    "pattern_count" / Int32ul,
    "is_absolute_path" / Int32ul,
    "vslib_version" / Int32ul,
    "sample_rate" / Int32ul,
    "numerator" / Int32ul,
    "denominator" / Int32ul,
    "padding1" / Bytes(0x4),
    "tempo" / Float64l,
    "master_volume" / Float64l,
    "is_float" / Int32ul,
    "bit_depth" / Int32ul,
    "channels" / Int32ul,
    "reserved" / Bytes(196),
)

VocalShifterProjectData = Struct(
    "magic" / Const(b"VSPD"),
    "version" / Bytes(8),
    "size" / Int32ul,
    "project_metadata" / VocalShifterProjectMetadata,
    "track_metadatas" / VocalShifterTrackMetadata[this.project_metadata.track_count],
    "pattern_metadatas" / VocalShifterPatternMetadata[this.project_metadata.pattern_count],
    "pattern_datas" / VocalShifterPatternData[this.project_metadata.pattern_count],
)



import io
from typing import TYPE_CHECKING
from xml.sax import saxutils

if TYPE_CHECKING:
    from collections.abc import Callable


class EchoGenerator(saxutils.XMLGenerator):
    # from https://code.activestate.com/recipes/84516-using-the-sax2-lexicalhandler-interface/

    def __init__(
        self,
        out: io.IOBase | None = None,
        encoding: str = "iso-8859-1",
        short_empty_elements: bool = False,
    ) -> None:
        super().__init__(out, encoding, short_empty_elements)
        self._in_cdata = 0
        self._write: Callable[[str], None]
        self._finish_pending_start_element: Callable[[], None]

    def characters(self, content: str) -> None:
        if self._in_cdata:
            self._write(content)
        else:
            super().characters(content)

    def start_cdata(self) -> None:
        self._finish_pending_start_element()
        self._write("<![CDATA[")
        self._in_cdata = 1

    def end_cdata(self) -> None:
        self._write("]]>")
        self._in_cdata = 0



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.core.constants import DEFAULT_BPM, TICKS_IN_BEAT
from libresvip.model.option_mixins import (
    EnablePitchImportationMixin,
    EnableVolumeImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class MultiChannelOption(Enum):
    FIRST: Annotated[str, Field(title=_("Import first channel only"))] = "first"
    SPLIT: Annotated[str, Field(title=_("Split into tracks"))] = "split"
    CUSTOM: Annotated[str, Field(title=_("Custom import range"))] = "custom"


class InputOptions(EnablePitchImportationMixin, EnableVolumeImportationMixin, BaseModel):
    import_lyrics: bool = Field(default=True, title=_("Import lyrics"))
    lyric_encoding: str = Field(
        default="utf-8",
        title=_("Lyric text encoding"),
        description=_("Unless the lyrics are garbled, this option should not be changed."),
    )
    import_time_signatures: bool = Field(
        default=True,
        title=_("Import time signatures"),
        description=_("If this option is unset, the time signature is set to 4/4."),
    )
    multi_channel: MultiChannelOption = Field(
        default=MultiChannelOption.FIRST,
        title=_("Multi-channel processing method"),
    )
    channels: str = Field(
        default="1",
        title=_("Channels to import"),
        description=_(
            'Specify which channels to import notes from. Enter channel numbers and/or channel ranges (separated by commas), e.g. 1,3,5-12. Range: 1-16. Only valid when the option "Custom import range" is selected in "Multi-channel processing method".'
        ),
    )
    default_bpm: float = Field(
        default=DEFAULT_BPM,
        title=_("Default BPM"),
        description=_("Used when no BPM information is found in the MIDI file."),
    )


class OutputOptions(BaseModel):
    export_lyrics: bool = Field(default=True, title=_("Export lyrics"))
    remove_symbols: bool = Field(
        default=True,
        title=_("Remove symbols from lyrics"),
        description=_(
            "Remove commas, periods, question marks and exclamation marks in Chinese and English to prevent lyric import failure in some vocal synthesizers."
        ),
    )
    compatible_lyric: bool = Field(
        default=False,
        title=_("Lyric compatibility mode"),
        description=_(
            "Convert all Chinese lyrics to pinyin to prevent garbled characters in MIDI files with Chinese lyrics from being imported by vocal synthesizers that do not support Chinese lyrics."
        ),
    )
    lyric_encoding: str = Field(
        default="utf-8",
        title=_("Lyric text encoding"),
        description=_("Unless the lyrics are garbled, this option should not be changed."),
    )
    transpose: int = Field(
        default=0,
        title=_("Transpose"),
    )
    ticks_per_beat: int = Field(
        default=TICKS_IN_BEAT,
        title=_("Ticks per beat"),
        description=_(
            "Also known as parts per quarter, ticks per quarter note, the number of pulses per quarter note. This setting should not be changed unless you know what it is."
        ),
    )



import abc
import pathlib
from typing import TypeVar

from libresvip.model.base import BaseModel, Project

BasePlugin_co = TypeVar("BasePlugin_co", bound="BasePlugin", covariant=True)


class BasePlugin(abc.ABC):
    pass


class SVSConverterBase(BasePlugin):
    @abc.abstractmethod
    def load(self, path: pathlib.Path, options: BaseModel) -> Project: ...

    @abc.abstractmethod
    def dump(self, path: pathlib.Path, project: Project, options: BaseModel) -> None: ...


class MiddlewareBase(BasePlugin):
    @abc.abstractmethod
    def process(self, project: Project, options: BaseModel) -> Project: ...


class WriteOnlyConverterBase(SVSConverterBase, abc.ABC):
    def load(self, path: pathlib.Path, options: BaseModel) -> Project:
        raise NotImplementedError


class ReadOnlyConverterBase(SVSConverterBase, abc.ABC):
    def dump(self, path: pathlib.Path, project: Project, options: BaseModel) -> None:
        raise NotImplementedError



import dataclasses
import math
import operator
import re
from collections.abc import Callable

from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.tick_counter import shift_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import clamp

from .base_pitch_curve import BasePitchCurve
from .enums import AcepLyricsLanguage
from .model import (
    AcepAudioTrack,
    AcepNote,
    AcepParamCurveList,
    AcepParams,
    AcepProject,
    AcepTempo,
    AcepTimeSignature,
    AcepTrack,
    AcepVocalTrack,
)
from .options import InputOptions, NormalizationMethod
from .singers import id2singer

ACEP_LATIN_SPAN_RE = re.compile(r"#(\d+)$")


@dataclasses.dataclass
class AceParser:
    options: InputOptions
    content_version: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_ticks: int = dataclasses.field(init=False)

    def parse_project(self, ace_project: AcepProject) -> Project:
        project = Project()
        self.content_version = ace_project.version
        project.time_signature_list = self.parse_time_signatures(ace_project.time_signatures)
        self.first_bar_ticks = int(project.time_signature_list[0].bar_length())
        project.song_tempo_list = shift_tempo_list(
            self.parse_tempos(ace_project.tempos),
            self.first_bar_ticks,
        )
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        for ace_track in ace_project.tracks:
            ace_track.gain += ace_project.master.gain
        for ace_track in ace_project.tracks:
            if (track := self.parse_track(ace_track)) is not None:
                project.track_list.append(track)
        return project

    @staticmethod
    def parse_time_signatures(ace_time_sigs: list[AcepTimeSignature]) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=ace_time_sig.bar_pos,
                numerator=ace_time_sig.numerator,
                denominator=ace_time_sig.denominator,
            )
            for ace_time_sig in ace_time_sigs
        ]

    @staticmethod
    def parse_tempos(ace_tempos: list[AcepTempo]) -> list[SongTempo]:
        return [
            SongTempo(position=ace_tempo.position, bpm=ace_tempo.bpm) for ace_tempo in ace_tempos
        ]

    def parse_track(self, ace_track: AcepTrack) -> Track | None:
        if (
            self.options.import_instrumental_track
            and isinstance(ace_track, AcepAudioTrack)
            and len(ace_track.patterns)
        ):
            track = InstrumentalTrack(
                audio_file_path=ace_track.patterns[0].path,
                offset=int(
                    ace_track.patterns[0].pos
                    if self.content_version < 7
                    else self.synchronizer.get_actual_ticks_from_secs(ace_track.patterns[0].pos)
                ),
            )
        elif isinstance(ace_track, AcepVocalTrack):
            track = SingingTrack(
                ai_singer_name=(id2singer.get(ace_track.singers[0].singer.singer_id, None) or "")
            )
            ace_note_list = []
            ace_params = AcepParams()
            for pattern in sorted(ace_track.patterns, key=operator.attrgetter("clip_pos")):
                ace_notes = [
                    note
                    for note in pattern.notes
                    if note.pos + pattern.pos >= 0
                    and pattern.clip_pos <= note.pos < pattern.clip_pos + pattern.clip_dur
                ]
                if not ace_notes:
                    continue
                prev_ace_note = None
                for ace_note in ace_notes:
                    ace_note.dur = int(
                        min(
                            ace_note.dur,
                            pattern.clip_pos + pattern.clip_dur - ace_note.pos,
                        )
                    )
                    ace_note.pos += int(pattern.pos)
                    if (
                        prev_ace_note is not None
                        and prev_ace_note.pos + prev_ace_note.dur > ace_note.pos
                    ):
                        prev_ace_note.dur = ace_note.pos - prev_ace_note.pos
                    prev_ace_note = ace_note
                ace_note_list.extend(ace_notes)

                def merge_curves(src: AcepParamCurveList, dst: AcepParamCurveList) -> None:
                    ace_curves = [
                        curve
                        for curve in src.root
                        if curve.offset + pattern.pos >= -self.first_bar_ticks
                        and curve.offset + len(curve.values) > pattern.clip_pos
                        and curve.offset < pattern.clip_pos + pattern.clip_dur
                    ]
                    for ace_curve in ace_curves:
                        max_length = int(pattern.clip_pos + pattern.clip_dur - ace_curve.offset)
                        if max_length < len(ace_curve.values):
                            ace_curve.values = ace_curve.values[:max_length]
                        ace_curve.offset += int(pattern.pos)
                    dst.root.extend(ace_curves)

                merge_curves(pattern.parameters.pitch_delta, ace_params.pitch_delta)
                merge_curves(pattern.parameters.breathiness, ace_params.breathiness)
                merge_curves(pattern.parameters.gender, ace_params.gender)
                merge_curves(pattern.parameters.energy, ace_params.energy)
                merge_curves(pattern.parameters.tension, ace_params.tension)
                if self.options.breath_normalization.enabled:
                    merge_curves(
                        pattern.parameters.real_breathiness,
                        ace_params.real_breathiness,
                    )
                if self.options.tension_normalization.enabled:
                    merge_curves(
                        pattern.parameters.real_tension,
                        ace_params.real_tension,
                    )
                if self.options.energy_normalization.enabled:
                    merge_curves(pattern.parameters.real_energy, ace_params.real_energy)
            ace_note_list.sort(key=operator.attrgetter("pos"))
            ace_params.pitch_delta.root.sort(key=operator.attrgetter("offset"))
            ace_params.breathiness.root.sort(key=operator.attrgetter("offset"))
            ace_params.gender.root.sort(key=operator.attrgetter("offset"))
            ace_params.energy.root.sort(key=operator.attrgetter("offset"))
            ace_params.tension.root.sort(key=operator.attrgetter("offset"))
            if self.options.breath_normalization.enabled:
                ace_params.real_breathiness.root.sort(key=operator.attrgetter("offset"))
            if self.options.tension_normalization.enabled:
                ace_params.real_tension.root.sort(key=operator.attrgetter("offset"))
            if self.options.energy_normalization.enabled:
                ace_params.real_energy.root.sort(key=operator.attrgetter("offset"))
            track.note_list = [self.parse_note(ace_note) for ace_note in ace_note_list]
            track.edited_params = self.parse_params(ace_params, ace_note_list)
        else:
            return None
        track.title = ace_track.name
        track.mute = ace_track.mute
        track.solo = ace_track.solo
        track.volume = 10 ** (ace_track.gain / 20)
        return track

    def parse_note(self, ace_note: AcepNote) -> Note:
        if (
            not self.options.keep_all_pronunciation
            and ace_note.language == AcepLyricsLanguage.CHINESE
        ):
            pronunciation = next(iter(get_pinyin_series(ace_note.lyric)), None)
        else:
            pronunciation = None
        note = Note(
            key_number=ace_note.pitch,
            start_pos=ace_note.pos,
            length=ace_note.dur,
            lyric=ace_note.lyric,
        )
        if (
            ace_note.language in [AcepLyricsLanguage.ENGLISH, AcepLyricsLanguage.SPANISH]
            and (latin_span := ACEP_LATIN_SPAN_RE.search(note.lyric)) is not None
        ):
            span_index = int(latin_span.group(1))
            note.lyric = ACEP_LATIN_SPAN_RE.sub("", note.lyric) if span_index == 1 else "+"
        if ace_note.syllable and ace_note.syllable != ace_note.freezed_default_syllable:
            note.pronunciation = ace_note.syllable
        elif pronunciation is None or (
            "-" not in ace_note.lyric and ace_note.pronunciation != pronunciation
        ):
            note.pronunciation = ace_note.pronunciation
        if ace_note.br_len > 0:
            note.head_tag = "V"
        if ace_note.head_consonants:
            note.edited_phones = Phones(
                head_length_in_secs=self.synchronizer.get_duration_secs_from_ticks(
                    note.start_pos - int(ace_note.head_consonants[0]),
                    note.start_pos,
                )
                if self.content_version < 7
                else ace_note.head_consonants[0]
            )
        return note

    def parse_params(self, ace_params: AcepParams, ace_note_list: list[AcepNote]) -> Params:
        def linear_transform(
            lower_bound: float, middle_value: float, upper_bound: float
        ) -> Callable[[float], int]:
            return lambda x: round(
                clamp(
                    (x - middle_value) / (upper_bound - middle_value) * 1000
                    if x >= middle_value
                    else (x - middle_value) / (middle_value - lower_bound) * 1000,
                    -1000,
                    1000,
                )
            )

        if self.options.breath_normalization.enabled:
            normalized = ace_params.real_breathiness.exclude(
                lambda x: (
                    x + 1e-3 < self.options.breath_normalization.lower_threshold
                    or x - 1e-3 > self.options.breath_normalization.upper_threshold
                )
            )
            if self.options.breath_normalization.normalize_method == NormalizationMethod.ZSCORE:
                normalized = normalized.z_score_normalize(
                    self.options.breath_normalization.scale,
                    self.options.breath_normalization.bias,
                )
            elif self.options.breath_normalization.normalize_method == NormalizationMethod.MINMAX:
                normalized = normalized.minmax_normalize(
                    self.options.breath_normalization.scale,
                    self.options.breath_normalization.bias,
                )
            else:
                msg = f"Unknown normalization method: {self.options.breath_normalization.normalize_method}"
                show_warning(msg)

            ace_params.breathiness = ace_params.breathiness.plus(
                normalized, 1.0, lambda x: x * 1.5 if x >= 0 else x * 0.8
            )

        if self.options.tension_normalization.enabled:
            normalized = ace_params.real_tension.exclude(
                lambda x: (
                    x + 1e-3 < self.options.tension_normalization.lower_threshold
                    or x - 1e-3 > self.options.tension_normalization.upper_threshold
                )
            )
            if self.options.tension_normalization.normalize_method == NormalizationMethod.ZSCORE:
                normalized = normalized.z_score_normalize(
                    self.options.tension_normalization.scale,
                    self.options.tension_normalization.bias,
                )
            elif self.options.tension_normalization.normalize_method == NormalizationMethod.MINMAX:
                normalized = normalized.minmax_normalize(
                    self.options.tension_normalization.scale,
                    self.options.tension_normalization.bias,
                )
            else:
                msg = f"Unknown normalization method: {self.options.tension_normalization.normalize_method}"
                show_warning(msg)

            ace_params.tension = ace_params.tension.plus(
                normalized, 1.0, lambda x: x * 0.5 if x >= 0 else x * 0.3
            )

        if self.options.energy_normalization.enabled:
            normalized = ace_params.real_energy.exclude(
                lambda x: (
                    x + 1e-3 < self.options.energy_normalization.lower_threshold
                    or x - 1e-3 > self.options.energy_normalization.upper_threshold
                )
            )
            if self.options.energy_normalization.normalize_method == NormalizationMethod.ZSCORE:
                normalized = normalized.z_score_normalize(
                    self.options.energy_normalization.scale,
                    self.options.energy_normalization.bias,
                )
            elif self.options.energy_normalization.normalize_method == NormalizationMethod.MINMAX:
                normalized = normalized.minmax_normalize(
                    self.options.energy_normalization.scale,
                    self.options.energy_normalization.bias,
                )
            else:
                msg = f"Unknown normalization method: {self.options.energy_normalization.normalize_method}"
                show_warning(msg)

            ace_params.energy = ace_params.energy.plus(normalized, 1.0, lambda x: x)

        parameters = Params()
        if self.options.import_pitch:
            parameters.pitch = self.parse_pitch_curve(ace_params.pitch_delta, ace_note_list)
        if self.options.import_breath:
            parameters.breath = self.parse_param_curve(
                ace_params.breathiness, linear_transform(0.2, 1, 2.5)
            )
        if self.options.import_gender:
            parameters.gender = self.parse_param_curve(
                ace_params.gender, linear_transform(-1, 0, 1)
            )
        if self.options.import_tension and self.options.import_energy:
            transform = linear_transform(0, 1, 2)
            parameters.volume = self.parse_param_curve(
                ace_params.energy,
                lambda x: round(self.options.energy_coefficient * transform(x)),
            )
            remaining_energy = ace_params.energy.model_copy(
                deep=True,
                update={
                    "root": [
                        part.model_copy(
                            deep=True,
                            update={
                                "values": [
                                    (value - 1) * (1 - self.options.energy_coefficient) + 1
                                    for value in part.values
                                ]
                            },
                        )
                        for part in ace_params.energy.root
                    ]
                },
            )
            energy_plus_tension = remaining_energy.plus(
                ace_params.tension,
                1.0,
                lambda x: (x - 1) * 0.5 if x >= 1 else (x - 1) * 0.3,
            )
            parameters.strength = self.parse_param_curve(
                energy_plus_tension,
                lambda x: round(self.options.energy_coefficient * transform(x)),
            )
        elif self.options.import_tension:
            parameters.strength = self.parse_param_curve(
                ace_params.tension, linear_transform(0.7, 1, 1.5)
            )
        elif self.options.import_energy:
            transform = linear_transform(0, 1, 2)
            parameters.volume = self.parse_param_curve(
                ace_params.energy,
                lambda x: round(self.options.energy_coefficient * transform(x)),
            )
            parameters.strength = self.parse_param_curve(
                ace_params.energy,
                lambda x: round((1 - self.options.energy_coefficient) * transform(x)),
            )
        return parameters

    def parse_pitch_curve(
        self, ace_curves: AcepParamCurveList, ace_note_list: list[AcepNote]
    ) -> ParamCurve:
        curve = ParamCurve()
        curve.points.append(Point.start_point())
        if len(ace_curves.root) > 0:
            base_pitch = BasePitchCurve(ace_note_list, self.synchronizer)
            for ace_curve in ace_curves.root:
                pos = ace_curve.offset
                curve.points.append(Point(pos + self.first_bar_ticks, -100))
                for value in ace_curve.values:
                    if ace_curve.curve_type == "anchor":
                        curve.points.append(Point(pos + self.first_bar_ticks, round(value * 100)))
                    elif not math.isnan(value):
                        abs_semitone = (
                            base_pitch.semitone_value_at(
                                self.synchronizer.get_actual_secs_from_ticks(pos)
                            )
                            + value
                        )
                        curve.points.append(
                            Point(
                                pos + self.first_bar_ticks,
                                round(abs_semitone * 100),
                            )
                        )
                    pos += 1
                curve.points.append(Point(pos - 1 + self.first_bar_ticks, -100))
        curve.points.append(Point.end_point())
        if self.options.curve_sample_interval > 0:
            curve = curve.reduce_sample_rate(self.options.curve_sample_interval, -100)
        return curve

    def parse_param_curve(
        self,
        ace_curves: AcepParamCurveList,
        mapping_func: Callable[[float], int],
    ) -> ParamCurve:
        curve = ParamCurve()
        curve.points.append(Point.start_point(0))
        for ace_curve in ace_curves.root:
            pos = ace_curve.offset
            curve.points.append(Point(x=pos + self.first_bar_ticks, y=0))
            for value in ace_curve.values:
                curve.points.append(Point(x=pos + self.first_bar_ticks, y=mapping_func(value)))
                pos += 1
            curve.points.append(Point(x=pos - 1 + self.first_bar_ticks, y=0))
        curve.points.append(Point.end_point(0))
        if self.options.curve_sample_interval > 0:
            curve = curve.reduce_sample_rate(self.options.curve_sample_interval)
        return curve



import contextlib
import encodings
import functools
import importlib
import pkgutil
import re
import textwrap
import uuid
from collections.abc import Callable
from typing import Any

import charset_normalizer.constant
import zhon
from retrie.retrie import WORD_BOUNDARY, Blacklist

LATIN_ALPHABET: re.Pattern[str] = re.compile(r"[a-zA-Z]+")
SYMBOL_PATTERN: re.Pattern[str] = re.compile(
    rf"(?!-)[\!\"\#\$%\&'\(\)\*,\./:;<=>\?\[\\\]\^_`\{{\|\}}\~{zhon.hanzi.punctuation}]+"
)


class CustomBoundriesMixin:
    re_flags: int
    parse_re_flags: Callable[[int], int]
    pattern: Callable[[], str]

    def __init__(
        self,
        *args: Any,
        left_boundary: str | None = None,
        right_boundary: str | None = None,
        **kwargs: Any,
    ) -> None:
        default_word_boundary = "" if kwargs.get("match_substrings") is True else WORD_BOUNDARY
        if left_boundary is None:
            left_boundary = default_word_boundary
        if right_boundary is None:
            right_boundary = default_word_boundary
        self.left_boundary = left_boundary
        self.right_boundary = right_boundary
        super().__init__(*args, **kwargs)

    def compile(
        self,
        word_boundary: str | None = None,
        re_flags: int = -1,
    ) -> re.Pattern[str]:
        left_boundary = self.left_boundary if word_boundary is None else word_boundary
        right_boundary = self.right_boundary if word_boundary is None else word_boundary
        re_flags = self.re_flags if re_flags == -1 else re_flags

        if left_boundary == WORD_BOUNDARY:
            lookbehind = left_boundary
        else:
            lookbehind = f"(?<={left_boundary})" if left_boundary else ""

        if right_boundary == WORD_BOUNDARY:
            lookahead = right_boundary
        else:
            lookahead = f"(?={right_boundary})" if right_boundary else ""

        return re.compile(
            lookbehind + self.pattern() + lookahead,
            flags=self.parse_re_flags(re_flags),
        )


class CustomBoundriesBlacklist(CustomBoundriesMixin, Blacklist):
    pass


def to_unicode(content: bytes) -> str:
    guessed_charset = charset_normalizer.detect(content)
    encoding = guessed_charset["encoding"] or "utf-8"
    return content.decode(encoding)


@functools.cache
def supported_charset_names() -> list[str]:
    encoding_names = []
    for module_info in pkgutil.walk_packages(encodings.__path__):
        cp_name = module_info.name
        with contextlib.suppress(ImportError):
            sub_module = importlib.import_module(f"encodings.{cp_name}")
            if not cp_name.endswith("_codec") and hasattr(sub_module, "getregentry"):
                with contextlib.suppress(ValueError):
                    iana_name = charset_normalizer.utils.iana_name(cp_name)
                    encoding_names.append(
                        charset_normalizer.constant.CHARDET_CORRESPONDENCE.get(iana_name, iana_name)
                    )
    return sorted(encoding_names)


def shorten_error_message(message: str | None) -> str:
    if message is None:
        return ""
    error_lines = textwrap.wrap(message, 70)
    if len(error_lines) > 30:
        message = "\n".join(error_lines[:15] + ["..."] + error_lines[-15:])
    else:
        message = "\n".join(error_lines)
    return message


def uuid_str() -> str:
    return str(uuid.uuid4())



from construct import (
    Check,
    Computed,
    Const,
    ExprAdapter,
    GreedyRange,
    If,
    IfThenElse,
    Int8ub,
    Int16ub,
    Int32ub,
    PaddedString,
    Select,
    Struct,
    obj_,
)
from pydantic import Field

from libresvip.model.base import BaseModel

# Reference: https://github.com/jazz-soft/JZZ/blob/master/javascript/JZZ.js
BASE_DCTPQ = 0x30
BASE_DELTA_TICKS = 0x40
BASE_NOTE = 0x7F
BASE_START_PITCH = 0x4090
BASE_END_PITCH = 0x4080

PPQN = ExprAdapter(
    Int32ub,
    encoder=obj_ + (BASE_DCTPQ << 16),
    decoder=obj_ - (BASE_DCTPQ << 16),
)

DeltaTicks = ExprAdapter(
    Int32ub,
    encoder=obj_ + (BASE_DELTA_TICKS << 16),
    decoder=obj_ - (BASE_DELTA_TICKS << 16),
)

NoteNumber = ExprAdapter(
    Int32ub,
    encoder=obj_ + (BASE_NOTE << 16),
    decoder=obj_ - (BASE_NOTE << 16),
)

StartPitch = ExprAdapter(
    Int32ub,
    encoder=(1 + (obj_ << 8) + (BASE_START_PITCH << 16)),
    decoder=(obj_ - (BASE_START_PITCH << 16)) >> 8,
)

EndPitch = ExprAdapter(
    Int32ub,
    encoder=(1 + (obj_ << 8) + (BASE_END_PITCH << 16)),
    decoder=(obj_ - (BASE_END_PITCH << 16)) >> 8,
)

UmpClipName = Struct(
    "header" / Const(b"\xd0"),
    "seq_stat" / Int8ub,
    "magic" / Const(b"\x01\x03"),
    "name" / PaddedString(12, "utf-8"),
    "type" / Computed("track_name"),
)

UmpStartOfClip = Struct(
    "magic" / Const(b"\xf0\x20\x00\x00"),
    "padding" / Const(b"\x00" * 12),
)

UmpEndOfClip = Struct(
    "magic" / Const(b"\xf0\x21\x00\x00"),
    "padding" / Const(b"\x00" * 12),
)

UmpTempo = Struct(
    "time" / DeltaTicks,
    "magic" / Const(b"\xd0\x10\x00\x00"),
    "tempo"
    / ExprAdapter(
        Int32ub,
        encoder=lambda obj, context: int(1705032600 / obj),
        decoder=lambda obj, context: 1705032600 / obj,
    ),
    "padding" / Const(b"\x00" * 8),
    "type" / Computed("set_tempo"),
)

UmpTimeSignature = Struct(
    "time" / DeltaTicks,
    "magic" / Const(b"\xd0\x10\x00\x01"),
    "numerator" / Int8ub,
    "denominator"
    / ExprAdapter(
        Int8ub,
        encoder=lambda obj, context: obj.bit_length() - 1,
        decoder=lambda obj, context: 1 << obj,
    ),
    "padding" / Const(b"\x08" + b"\x00" * 9),
    "type" / Computed("time_signature"),
)

UmpLyric = Struct(
    "time" / DeltaTicks,
    "header" / Const(b"\xd0"),
    "seq_stat" / Int8ub,
    "magic" / Const(b"\x01\x50"),
    "seq_num" / If(lambda this: (this.seq_stat - 16) // 64 <= 1, Int16ub),
    "text"
    / IfThenElse(
        lambda this: (this.seq_stat - 16) // 64 <= 1,
        PaddedString(10, "ascii"),
        PaddedString(12, "ascii"),
    ),
    "type" / Computed("lyrics"),
)

UmpMetadata = Struct(
    "time" / DeltaTicks,
    "header" / Const(b"\xd0"),
    "seq_stat" / Int8ub,
    "magic" / Const(b"\x01\x51"),
    "text" / PaddedString(12, "utf-8"),
    "type" / Computed("metadata"),
)

UmpNoteOn = Struct(
    "time" / DeltaTicks,
    "note" / StartPitch,
    Check(lambda ctx: 0 <= ctx.note <= 127),
    "velocity" / Int16ub,
    "on_data" / Int16ub,
    "type" / Computed("note_on"),
)

UmpNoteOff = Struct(
    "time" / DeltaTicks,
    "note" / EndPitch,
    Check(lambda ctx: 0 <= ctx.note <= 127),
    "velocity" / Int16ub,
    "off_data" / Int16ub,
    "type" / Computed("note_off"),
)

VxTrack = Struct(
    "title_parts" / GreedyRange(UmpClipName),
    "start_of_clip" / UmpStartOfClip,
    "events"
    / GreedyRange(Select(UmpTempo, UmpTimeSignature, UmpLyric, UmpMetadata, UmpNoteOn, UmpNoteOff)),
    "end_of_clip" / UmpEndOfClip,
)

VxFile = Struct(
    "magic" / Const(b"SMF2CLIP"),
    "ticks_per_beat" / PPQN,
    "tracks" / GreedyRange(VxTrack),
)


class VxPitchPoint(BaseModel):
    pitch: float
    position: int
    applicable: bool = True


class VxTimeBasedPitchSequence(BaseModel):
    time_frame_period_seconds: float = Field(alias="timeFramePeriodSeconds")
    num_frames_overall_sequence: int = Field(0, alias="numFramesOverallSequence")
    pitch_sequence: list[VxPitchPoint] = Field(default_factory=list, alias="pitchSequence")


class VxPitchData(BaseModel):
    time_based_pitch_sequence: VxTimeBasedPitchSequence = Field(alias="TimeBasedPitchSequence")



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    default_singer_name: str = Field("さとうささら", title=_("Default Singer Name"))
    default_singer_id: str = Field("A", title=_("Default Voicebank ID"))
    default_singer_version: str = Field("1.0.0", title=_("Default Voicebank Version"))



import random

_colors = [
    "#24936E",
    "#B54434",
    "#B5495B",
    "#A8497A",
    "#4F726C",
    "#939650",
    "#CA7A2C",
    "#DB4D6D",
    "#77428D",
    "#005CAF",
]


def count_color() -> int:
    return len(_colors)


def get_color(index: int) -> str:
    return _colors[index % len(_colors)]


def random_color() -> str:
    return random.choice(_colors)



import dataclasses
import functools
import math
import pathlib

import more_itertools
import portion

from libresvip.core.constants import DEFAULT_ENGLISH_LYRIC
from libresvip.core.tick_counter import skip_beat_list, skip_tempo_list
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .constants import BPM_RATE
from .model import (
    Vsq3ParameterNames,
    Vsq4ParameterNames,
    Vsqx,
    VsqxMasterTrack,
    VsqxMusicalPart,
    VsqxNote,
    VsqxParameterNames,
    VsqxTempoList,
    VsqxTimeSigList,
    VsqxVsTrackList,
    VsqxVsUnitList,
    VsqxVVoice,
    VsqxWavPartList,
    VsqxWavUnitList,
)
from .options import InputOptions
from .vocaloid_pitch import (
    ControllerEvent,
    VocaloidPartPitchData,
    pitch_from_vocaloid_parts,
)


@dataclasses.dataclass
class VsqxParser:
    options: InputOptions
    src_path: pathlib.Path
    param_names: type[VsqxParameterNames] = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    pc2voice: dict[int, VsqxVVoice] = dataclasses.field(default_factory=dict)

    def parse_project(self, vsqx_project: Vsqx) -> Project:
        if vsqx_project.Meta.name == "vsq3":
            self.param_names = Vsq3ParameterNames
        elif vsqx_project.Meta.name == "vsq4":
            self.param_names = Vsq4ParameterNames
        master_track: VsqxMasterTrack = vsqx_project.master_track
        tick_prefix, time_signatures = self.parse_time_signatures(
            master_track.time_sig, master_track.pre_measure
        )
        tempos = self.parse_tempos(master_track.tempo, tick_prefix)
        self.synchronizer = TimeSynchronizer(tempos)
        for v_voice in vsqx_project.v_voice_table.v_voice:
            if v_voice.v_pc is not None:
                self.pc2voice[v_voice.v_pc] = v_voice
        singing_tracks = self.parse_singing_tracks(
            vsqx_project.vs_track, vsqx_project.mixer.vs_unit, tick_prefix
        )
        wav_parts: VsqxWavPartList = []
        wav_units: VsqxWavUnitList = []
        if vsqx_project.mono_track is not None:
            wav_parts += vsqx_project.mono_track.wav_part
            wav_units += [vsqx_project.mixer.mono_unit] * len(vsqx_project.mono_track.wav_part)
        if vsqx_project.stereo_track is not None:
            wav_parts += vsqx_project.stereo_track.wav_part
            wav_units += [vsqx_project.mixer.stereo_unit] * len(vsqx_project.stereo_track.wav_part)
        instrumental_tracks = self.parse_instrumental_tracks(wav_parts, wav_units, tick_prefix)
        return Project(
            song_tempo_list=tempos,
            time_signature_list=time_signatures,
            track_list=singing_tracks + instrumental_tracks,
        )

    def parse_time_signatures(
        self, time_signatures: VsqxTimeSigList, measure_prefix: int
    ) -> tuple[int, list[TimeSignature]]:
        time_signature_list = [
            TimeSignature(
                numerator=time_sig.nume,
                denominator=time_sig.denomi,
                bar_index=time_sig.pos_mes,
            )
            for time_sig in time_signatures
        ]
        tick_prefix = 0
        measure = 0
        for time_sig in time_signature_list:
            measure_diff = time_sig.bar_index - measure
            tick_prefix += measure_diff * round(time_sig.bar_length())
            measure += time_sig.bar_index
        measure_diff = measure_prefix - measure
        tick_prefix += measure_diff * round(time_signature_list[-1].bar_length())
        self.first_bar_length = round(time_signature_list[0].bar_length())
        return int(tick_prefix), skip_beat_list(time_signature_list, measure_prefix)

    def parse_tempos(self, tempos: VsqxTempoList, tick_prefix: int) -> list[SongTempo]:
        tempo_list = [
            SongTempo(
                position=tempo.pos_tick,
                bpm=tempo.bpm / BPM_RATE,
            )
            for tempo in tempos
        ]
        return skip_tempo_list(tempo_list, tick_prefix)

    def parse_singing_tracks(
        self,
        vs_tracks: VsqxVsTrackList,
        vs_units: VsqxVsUnitList,
        tick_prefix: int,
    ) -> list[SingingTrack]:
        singing_tracks = []
        for vs_track, vs_unit in zip(vs_tracks, vs_units):
            singing_track = SingingTrack(
                title=vs_track.track_name, mute=vs_unit.mute, solo=vs_unit.solo
            )
            for musical_part in vs_track.musical_part:
                (
                    note_list,
                    vibrato_rate_interval_dict,
                    vibrato_depth_interval_dict,
                ) = self.parse_notes(musical_part.note, musical_part.pos_tick - tick_prefix)
                singing_track.note_list.extend(note_list)
                if (
                    not singing_track.ai_singer_name
                    and musical_part.singer
                    and (v_voice := self.pc2voice.get(musical_part.singer[0].v_pc))
                    and v_voice.v_voice_name is not None
                ):
                    singing_track.ai_singer_name = v_voice.v_voice_name
                if self.options.import_pitch and (
                    pitch := self.parse_pitch(
                        musical_part,
                        note_list,
                        vibrato_rate_interval_dict,
                        vibrato_depth_interval_dict,
                        tick_prefix,
                    )
                ):
                    singing_track.edited_params.pitch.points.extend(pitch.points)
            if len(singing_track.edited_params.pitch.points.root):
                singing_track.edited_params.pitch.points.root.insert(0, Point.start_point())
                singing_track.edited_params.pitch.points.root.append(Point.end_point())
            singing_tracks.append(singing_track)
        return singing_tracks

    @staticmethod
    def vibrato_curve(value: float, shift: float, omega: float, phase: float) -> float:
        return math.cos(omega * (value - shift) + phase)

    def parse_notes(
        self, vsqx_notes: list[VsqxNote], tick_offset: int
    ) -> tuple[list[Note], PiecewiseIntervalDict, PiecewiseIntervalDict]:
        prev_vsqx_note = None
        note_list: list[Note] = []
        vibrato_depth_interval_dict = PiecewiseIntervalDict()
        vibrato_rate_interval_dict = PiecewiseIntervalDict()
        for vsqx_note in vsqx_notes:
            if prev_vsqx_note and vsqx_note.lyric.startswith("EVEC("):
                note_list[-1].length += vsqx_note.dur_tick
                continue
            elif vsqx_note.phnms is None or vsqx_note.phnms.value not in [
                "Asp",
                "Sil",
                "?",
            ]:
                note = Note(
                    start_pos=vsqx_note.pos_tick + tick_offset,
                    length=vsqx_note.dur_tick,
                    lyric=(vsqx_note.lyric or DEFAULT_ENGLISH_LYRIC).lower(),
                    pronunciation=vsqx_note.phnms.value
                    if (vsqx_note.phnms is not None and vsqx_note.phnms.lock == 1)
                    else None,
                    key_number=vsqx_note.note_num,
                )
                if vsqx_note.note_style.seq_attr:
                    start_secs = self.synchronizer.get_actual_secs_from_ticks(note.start_pos)
                    duration_secs = self.synchronizer.get_duration_secs_from_ticks(
                        note.start_pos, note.end_pos
                    )
                    for seq_attr in vsqx_note.note_style.seq_attr:
                        if seq_attr.seq_id == "vibRate":
                            phase = 0
                            for prev_elem, elem in more_itertools.windowed(
                                [*seq_attr.elem, None], 2
                            ):
                                prev_start = start_secs + duration_secs * prev_elem.pos_nrm / 65536
                                omega = prev_elem.elv / 2
                                if elem is None:
                                    prev_end = start_secs + duration_secs
                                    vibrato_rate_interval_dict[
                                        portion.closed(prev_start, prev_end)
                                    ] = functools.partial(
                                        self.vibrato_curve,
                                        shift=prev_start,
                                        omega=omega,
                                        phase=phase,
                                    )
                                else:
                                    prev_end = start_secs + duration_secs * elem.pos_nrm / 65536
                                    vibrato_rate_interval_dict[
                                        portion.closedopen(prev_start, prev_end)
                                    ] = functools.partial(
                                        self.vibrato_curve,
                                        shift=prev_start,
                                        omega=omega,
                                        phase=phase,
                                    )
                                phase += (prev_end - prev_start) * omega
                        elif seq_attr.seq_id == "vibDep":
                            for prev_elem, elem in more_itertools.windowed(
                                [*seq_attr.elem, None], 2
                            ):
                                prev_start = start_secs + duration_secs * prev_elem.pos_nrm / 65536
                                if elem is None:
                                    prev_end = start_secs + duration_secs
                                    vibrato_depth_interval_dict[
                                        portion.closed(prev_start, prev_end)
                                    ] = prev_elem.elv
                                else:
                                    prev_end = start_secs + duration_secs * elem.pos_nrm / 65536
                                    vibrato_depth_interval_dict[
                                        portion.closedopen(prev_start, prev_end)
                                    ] = prev_elem.elv
                if (
                    prev_vsqx_note is not None
                    and prev_vsqx_note.phnms is not None
                    and prev_vsqx_note.pos_tick + prev_vsqx_note.dur_tick == vsqx_note.pos_tick
                    and prev_vsqx_note.phnms.value == "Sil"
                ):
                    note.head_tag = "0"
                note_list.append(note)
            prev_vsqx_note = vsqx_note
        if self.options.combine_syllables:
            for syllables_group in more_itertools.split_when(
                note_list,
                lambda prev_vsqx_note, vsqx_note: not prev_vsqx_note.lyric.endswith("-"),
            ):
                if len(syllables_group) > 1:
                    syllables_group[0].lyric = "".join(
                        vsqx_note.lyric.rstrip("-") for vsqx_note in syllables_group
                    )
                    for vsqx_note in syllables_group[1:]:
                        if vsqx_note.lyric != "-":
                            vsqx_note.lyric = "+"
        return (
            note_list,
            vibrato_rate_interval_dict,
            vibrato_depth_interval_dict,
        )

    def parse_pitch(
        self,
        musical_part: VsqxMusicalPart,
        note_list: list[Note],
        vibrato_rate_interval_dict: PiecewiseIntervalDict,
        vibrato_depth_interval_dict: PiecewiseIntervalDict,
        tick_offset: int,
    ) -> ParamCurve | None:
        pitch_data = VocaloidPartPitchData(
            start_pos=musical_part.pos_tick - tick_offset,
            pit=[
                ControllerEvent(
                    pos=music_control.pos_tick,
                    value=music_control.attr.value,
                )
                for music_control in musical_part.m_ctrl
                if music_control.attr.type_param_attr_id == self.param_names.PIT.value
                and music_control.attr.value is not None
            ],
            pbs=[
                ControllerEvent(
                    pos=music_control.pos_tick,
                    value=music_control.attr.value,
                )
                for music_control in musical_part.m_ctrl
                if music_control.attr.type_param_attr_id == self.param_names.PBS.value
                and music_control.attr.value is not None
            ],
        )
        return pitch_from_vocaloid_parts(
            [pitch_data],
            self.synchronizer,
            note_list,
            vibrato_rate_interval_dict,
            vibrato_depth_interval_dict,
            self.first_bar_length,
            musical_part.pos_tick - tick_offset,
            musical_part.pos_tick + musical_part.play_time - tick_offset,
        )

    def parse_instrumental_tracks(
        self,
        wav_parts: VsqxWavPartList,
        wav_units: VsqxWavUnitList,
        tick_prefix: int,
    ) -> list[InstrumentalTrack]:
        instrumental_tracks = []
        if self.options.import_instrumental_track:
            for wav_part, wav_unit in zip(wav_parts, wav_units):
                if pathlib.Path(wav_part.file_path).is_absolute():
                    wav_path_str = wav_part.file_path
                elif (
                    wav_path := self.src_path.with_suffix(".wavparts") / wav_part.file_path
                ).exists():
                    wav_path_str = str(wav_path)
                else:
                    continue
                instrumental_tracks.append(
                    InstrumentalTrack(
                        title=wav_part.part_name,
                        audio_file_path=wav_path_str,
                        offset=wav_part.pos_tick - tick_prefix,
                        mute=wav_unit.mute,
                        solo=wav_unit.solo,
                    )
                )
        return instrumental_tracks



from PyInstaller.utils.hooks import (  # noqa: N999
    collect_data_files,
    collect_submodules,
)

hiddenimports = (
    collect_submodules("libresvip.core")
    + collect_submodules("libresvip.model")
    + collect_submodules("libresvip.utils")
)

datas = (
    collect_data_files("libresvip.middlewares", include_py_files=True, excludes=["**/*.po"])
    + collect_data_files("libresvip.plugins", include_py_files=True, excludes=["**/*.po"])
    + collect_data_files(
        "libresvip.res", excludes=["**/*.po", "**/*.qml", "**/*.qrc", "**/*.ts", "**/qmldir"]
    )
)



import os
from multiprocessing import freeze_support

freeze_support()

from libresvip.core.constants import app_dir  # noqa: E402

os.environ.setdefault("NICEGUI_STORAGE_PATH", str(app_dir.user_config_path / ".nicegui"))



from importlib.resources import files
from typing import Final

import platformdirs

PACKAGE_NAME: Final[str] = "libresvip"

# lyric related constants
DEFAULT_PHONEME: Final[str] = "la"
DEFAULT_ENGLISH_LYRIC: Final[str] = DEFAULT_PHONEME
DEFAULT_SPANISH_LYRIC: Final[str] = DEFAULT_ENGLISH_LYRIC
DEFAULT_CHINESE_LYRIC: Final[str] = "啦"
DEFAULT_JAPANESE_LYRIC: Final[str] = "ラ"
DEFAULT_KOREAN_LYRIC: Final[str] = "라"

# midi related constants
DEFAULT_BPM: Final[float] = 120.0
KEY_IN_OCTAVE: Final[int] = 12
TICKS_IN_BEAT: Final[int] = 480
MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS: Final[int] = TICKS_IN_BEAT
DEFAULT_PITCH_BEND_SENSITIVITY: Final[int] = 2
MAX_PITCH_BEND_SENSITIVITY: Final[int] = 24
PITCH_MAX_VALUE: Final[int] = 8191

app_dir = platformdirs.AppDirs(PACKAGE_NAME)

pkg_dir = files(PACKAGE_NAME)
res_dir = pkg_dir / "res"



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class NoteLengthOption(Enum):
    ZERO: Annotated[str, Field(title=_("Zero length note"))] = "0/1"
    EIGHTH: Annotated[str, Field(title=_("1/8 note"))] = "1/8"
    SIXTEENTH: Annotated[str, Field(title=_("1/16 note"))] = "1/16"
    THIRTY_SECOND: Annotated[str, Field(title=_("1/32 note"))] = "1/32"
    SIXTY_FOURTH: Annotated[str, Field(title=_("1/64 note"))] = "1/64"
    ONE_HUNDRED_AND_TWENTY_EIGHTH: Annotated[
        str,
        Field(
            title=_("1/128 note"),
        ),
    ] = "1/128"


class ProcessOptions(BaseModel):
    fill_threshold: NoteLengthOption = Field(
        default=NoteLengthOption.ZERO,
        title=_("Max length to be processed (exclusive)"),
        description=_("Extend note to fill the short silences between it and its next note"),
    )



import dataclasses
import itertools
import operator

import more_itertools
from wanakana import PROLONGED_SOUND_MARK

from libresvip.core.tick_counter import (
    shift_beat_list,
    shift_tempo_list,
    skip_beat_list,
    skip_tempo_list,
)
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)

from .constants import OCTAVE_OFFSET, TICK_RATE
from .model import (
    VoiSonaMobilePointData,
    VoiSonaMobileProject,
    VoiSonaMobileStateInformation,
)
from .options import InputOptions
from .voisona_mobile_pitch import (
    VoiSonaMobileParamEvent,
    VoiSonaMobileTrackPitchData,
    pitch_from_voisona_track,
)


@dataclasses.dataclass
class VoiSonaMobileParser:
    options: InputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def parse_project(self, voisona_project: VoiSonaMobileProject) -> Project:
        time_signatures = []
        tempos = []
        tracks = []
        for track in voisona_project.mobile_singer:
            if parse_result := self.parse_singing_track(track.mobile_singer_data.state_information):
                singing_track, tempo_part, time_signature_part = parse_result
                tracks.append(singing_track)
                tempos.extend(tempo_part)
                time_signatures.extend(time_signature_part)
        tempos = self.merge_tempos(tempos)
        self.time_synchronizer = TimeSynchronizer(tempos)
        time_signatures = self.merge_time_signatures(time_signatures)
        return Project(
            time_signature_list=skip_beat_list(time_signatures, 0),
            song_tempo_list=skip_tempo_list(tempos, 0),
            track_list=tracks,
        )

    def merge_tempos(self, tempos: list[SongTempo]) -> list[SongTempo]:
        buckets = more_itertools.bucket(tempos, key=operator.attrgetter("position"))
        return [next(buckets[key]) for key in buckets] or [SongTempo()]

    def merge_time_signatures(self, time_signatures: list[TimeSignature]) -> list[TimeSignature]:
        buckets = more_itertools.bucket(time_signatures, key=operator.attrgetter("bar_index"))
        return [next(buckets[key]) for key in buckets] or [TimeSignature()]

    def parse_singing_track(
        self, state_information: VoiSonaMobileStateInformation
    ) -> tuple[SingingTrack, list[SongTempo], list[TimeSignature]] | None:
        if state_information.song is None:
            return None
        time_signatures = [
            TimeSignature(bar_index=0, numerator=4, denominator=4),
        ]
        prev_tick = 0
        tempos = []
        notes = []

        tick_prefix = int(time_signatures[0].bar_length())
        for song in state_information.song:
            for beat in song.beat:
                for time_node in beat.time:
                    tick = int(time_node.clock / TICK_RATE)
                    numerator = time_node.beats
                    denominator = time_node.beat_type

                    ticks_in_measure = time_signatures[-1].bar_length()
                    tick_diff = tick - prev_tick
                    measure_diff = tick_diff / ticks_in_measure
                    time_signatures.append(
                        TimeSignature(
                            bar_index=int(
                                time_signatures[-1].bar_index + measure_diff,
                            ),
                            numerator=numerator,
                            denominator=denominator,
                        )
                    )
                    prev_tick = tick
            for tempo in song.tempo:
                for tempo_node in tempo.sound:
                    tick = int(tempo_node.clock / TICK_RATE)
                    bpm = float(tempo_node.tempo)
                    tempos.append(SongTempo(position=tick, bpm=bpm))
            for score in song.score:
                for note_node in score.note:
                    pitch_octave = note_node.pitch_octave - OCTAVE_OFFSET
                    phoneme = None
                    if note_node.phoneme:
                        phoneme = note_node.phoneme.replace(",", " ")
                    notes.append(
                        Note(
                            key_number=note_node.pitch_step + pitch_octave * 12,
                            lyric="-"
                            if note_node.lyric == chr(PROLONGED_SOUND_MARK)
                            else note_node.lyric,
                            start_pos=(note_node.clock // TICK_RATE),
                            length=note_node.duration // TICK_RATE,
                            pronunciation=phoneme,
                        )
                    )
        tempos = shift_tempo_list(tempos, tick_prefix)
        voisona_track_pitch_data = None
        if state_information.parameter is not None:
            for parameter in state_information.parameter:
                if parameter.log_f0 is not None:
                    pitch_data_nodes = itertools.chain.from_iterable(
                        curve.data for curve in parameter.log_f0
                    )
                    vibrato_amplitude_nodes = itertools.chain.from_iterable(
                        curve.data for curve in parameter.vib_amp or []
                    )
                    vibrato_frequency_nodes = itertools.chain.from_iterable(
                        curve.data for curve in parameter.vib_frq or []
                    )
                    pitch_datas = [
                        pitch_data
                        for data_node in pitch_data_nodes
                        if (pitch_data := self.parse_param_data(data_node))
                    ]
                    vibrato_amplitude_data = [
                        vibrato_amplitude
                        for vibrato_amplitude_node in vibrato_amplitude_nodes
                        if (vibrato_amplitude := self.parse_param_data(vibrato_amplitude_node))
                    ]
                    vibrato_frequency_data = [
                        vibrato_frequency
                        for vibrato_frequency_node in vibrato_frequency_nodes
                        if (vibrato_frequency := self.parse_param_data(vibrato_frequency_node))
                    ]
                    voisona_track_pitch_data = VoiSonaMobileTrackPitchData(
                        events=pitch_datas,
                        tempos=tempos,
                        tick_prefix=tick_prefix,
                        vibrato_amplitude_events=vibrato_amplitude_data,
                        vibrato_frequency_events=vibrato_frequency_data,
                    )
        time_signatures = shift_beat_list(time_signatures, 1)
        singing_track = SingingTrack(note_list=notes)
        if (
            self.options.import_pitch
            and voisona_track_pitch_data is not None
            and (pitch := pitch_from_voisona_track(voisona_track_pitch_data)) is not None
        ):
            singing_track.edited_params.pitch = pitch
        return singing_track, tempos, time_signatures

    @staticmethod
    def parse_param_data(
        data_element: VoiSonaMobilePointData,
    ) -> VoiSonaMobileParamEvent | None:
        value = float(data_element.value)
        index = data_element.index or None
        repeat = data_element.repeat or None
        return VoiSonaMobileParamEvent(idx=index, repeat=repeat, value=value)



import dataclasses
import functools
import math
import re
from collections.abc import Callable
from typing import Concatenate

from more_itertools import pairwise
from typing_extensions import ParamSpec

from libresvip.core.constants import KEY_IN_OCTAVE
from libresvip.model.point import Point

P = ParamSpec("P")
NOTE_RE = re.compile(r"^(?P<note>[A-Ga-g])(?P<accidental>[#♯𝄪b!♭𝄫♮]*)(?P<octave>[+-]?\d+)?$")


def midi2note(midi: float) -> str:
    pitch_map = [
        "C",
        "C#",
        "D",
        "D#",
        "E",
        "F",
        "F#",
        "G",
        "G#",
        "A",
        "A#",
        "B",
    ]
    midi = round(midi)
    octave = (midi // KEY_IN_OCTAVE) - 1
    pitch = pitch_map[midi % KEY_IN_OCTAVE]
    return f"{pitch}{octave}"


def note2midi(note: str) -> int:
    pitch_map = {"C": 0, "D": 2, "E": 4, "F": 5, "G": 7, "A": 9, "B": 11}
    acc_map = {
        "#": 1,
        "": 0,
        "b": -1,
        "!": -1,
        "♯": 1,
        "𝄪": 2,
        "♭": -1,
        "𝄫": -2,
        "♮": 0,
    }

    if (match := NOTE_RE.match(note)) is None:
        msg = f"Invalid note format: {note!r}"
        raise ValueError(msg)

    pitch = match["note"].upper()
    offset = sum(acc_map[o] for o in match["accidental"])
    octave = match["octave"]

    octave = int(octave) if octave else 0

    note_value = KEY_IN_OCTAVE * (octave + 1) + pitch_map[pitch] + offset

    return round(note_value)


def hz2midi(hz: float, a4_midi: int = 69, base_freq: float = 440.0) -> float:
    return a4_midi + KEY_IN_OCTAVE * math.log2(hz / base_freq)


def midi2hz(midi: float, a4_midi: int = 69, base_freq: float = 440.0) -> float:
    return base_freq * 2 ** ((midi - a4_midi) / KEY_IN_OCTAVE)


def clamp(
    x: float,
    lower: float | None = None,
    upper: float | None = None,
) -> float:
    """Limit a value to a given range.

    The returned value is guaranteed to be between *lower* and
    *upper*. Integers, floats, and other comparable types can be
    mixed.

    Similar to `numpy's clip`_ function.

    .. _numpy's clip: http://docs.scipy.org/doc/numpy/reference/generated/numpy.clip.html
    .. from boltons: https://boltons.readthedocs.io/en/latest/mathutils.html#boltons.mathutils.clamp

    """
    if lower is None:
        lower = float("-inf")
    if upper is None:
        upper = float("inf")
    if upper < lower:
        msg = f"expected upper bound ({upper!r}) >= lower bound ({lower!r})"
        raise ValueError(msg)
    return min(max(x, lower), upper)


def _transform_interpolation_args(
    func: Callable[Concatenate[float, P], float],
) -> Callable[[float, tuple[float, float], tuple[float, float]], float]:
    @functools.wraps(func, assigned=["__module__", "__name__", "__qualname__", "__doc__"])
    def inner(
        x: float,
        start: tuple[float, float],
        end: tuple[float, float],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> float:
        x0, y0 = start
        x1, y1 = end
        r = (x - x0) / (x1 - x0)
        return y0 + (y1 - y0) * func(r, *args, **kwargs)

    return inner


@_transform_interpolation_args
def linear_interpolation(r: float) -> float:
    return r


@_transform_interpolation_args
def cosine_easing_in_interpolation(r: float) -> float:
    return 1 - math.cos(r * math.pi / 2)


@_transform_interpolation_args
def cosine_easing_out_interpolation(r: float) -> float:
    return math.sin(r * math.pi / 2)


@_transform_interpolation_args
def cosine_easing_in_out_interpolation(r: float) -> float:
    return (1 - math.cos(r * math.pi)) / 2


@_transform_interpolation_args
def cubic_interpolation(r: float) -> float:
    return (3 - 2 * r) * r**2


@_transform_interpolation_args
def vocaloid_interpolation(r: float) -> float:
    return math.sin(r * math.pi) / 2 if r <= 0.5 else r


@_transform_interpolation_args
def sigmoid_interpolation(r: float, k: float) -> float:
    return 1 / (1 + math.exp(k * (-2 * r + 1)))


def _inner_interpolate(
    data: list[Point],
    sampling_interval_tick: int,
    mapping: Callable[[int, Point, Point], float],
) -> list[Point]:
    return (
        (
            [data[0]]
            + [
                Point(x=x, y=round(mapping(x, start, end)))
                for start, end in pairwise(data)
                for x in range(start.x + 1, end.x, sampling_interval_tick)
            ]
            + [data[-1]]
        )
        if data
        else data
    )


interpolate_linear = functools.partial(_inner_interpolate, mapping=linear_interpolation)
interpolate_cosine_ease_in_out = functools.partial(
    _inner_interpolate, mapping=cosine_easing_in_out_interpolation
)
interpolate_cosine_ease_in = functools.partial(
    _inner_interpolate, mapping=cosine_easing_in_interpolation
)
interpolate_cosine_ease_out = functools.partial(
    _inner_interpolate, mapping=cosine_easing_out_interpolation
)


@dataclasses.dataclass
class HermiteInterpolator:
    # from https://github.com/LiuYunPlayer/TuneLab/blob/master/TuneLab.Base/Science/HermiteInterpolation.cs

    points: list[tuple[float, float]]

    @staticmethod
    def f1(t1: float, t2: float) -> float:
        return (1 + 2 * t1) * t2**2

    @staticmethod
    def f3(t: float, d: float) -> float:
        return t**2 * d

    @staticmethod
    def slope(p1: tuple[float, float], p2: tuple[float, float]) -> float:
        return (p2[1] - p1[1]) / (p2[0] - p1[0]) if p2[0] != p1[0] else math.nan

    def slope_at(self, point_index: int) -> float:
        if point_index in [0, len(self.points) - 1]:
            return 0
        point = self.points[point_index]
        last_k = self.slope(point, self.points[point_index - 1])
        next_k = self.slope(point, self.points[point_index + 1])
        kk = last_k * next_k
        return 0 if kk <= 0 else 2 / (1 / last_k + 1 / next_k)

    def interpolate(self, xs: list[float]) -> list[float]:
        if len(self.points) < 2:
            return [self.points[0][1] if len(self.points) == 1 else 0] * len(xs)
        elif len(self.points) == 2:
            return [linear_interpolation(x, self.points[0], self.points[1]) for x in xs]
        point_index = 0
        ys = []
        for x in xs:
            while point_index < len(self.points) and self.points[point_index][0] < x:
                point_index += 1
            if point_index == 0:
                ys.append(self.points[0][1])
            elif point_index == len(self.points):
                ys.append(self.points[-1][1])
            else:
                last_point = self.points[point_index - 1]
                last_delta = self.slope_at(point_index - 1)
                next_point = self.points[point_index]
                next_delta = self.slope_at(point_index)
                delta_1 = x - last_point[0]
                delta_2 = x - next_point[0]
                t1 = delta_1 / (next_point[0] - last_point[0])
                t2 = delta_2 / (last_point[0] - next_point[0])
                ys.append(
                    self.f1(t1, t2) * last_point[1]
                    + self.f1(t2, t1) * next_point[1]
                    + self.f3(t2, delta_1) * last_delta
                    + self.f3(t1, delta_2) * next_delta
                )
        return ys


def db_to_float(db: float, using_amplitude: bool = True) -> float:
    """
    Converts the input db to a float, which represents the equivalent
    ratio in power.
    """
    return 10 ** (db / 20) if using_amplitude else 10 ** (db / 10)


def ratio_to_db(ratio: float, val2: float | None = None, using_amplitude: bool = True) -> float:
    """
    Converts the input float to db, which represents the equivalent
    to the ratio in power represented by the multiplier passed in.
    """

    # accept 2 values and use the ratio of val1 to val2
    if val2 is not None:
        ratio /= val2

    # special case for multiply-by-zero (convert to silence)
    if ratio == 0:
        return -float("inf")

    return 20 * math.log10(ratio) if using_amplitude else 10 * math.log10(ratio)



import dataclasses

from libresvip.core.tick_counter import skip_beat_list
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import (
    MutaNote,
    MutaPoint,
    MutaProject,
    MutaTempo,
    MutaTimeSignature,
    MutaTrack,
    MutaTrackType,
)
from .options import InputOptions


@dataclasses.dataclass
class MutaParser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, muta_project: MutaProject) -> Project:
        time_signatures = self.parse_time_signatures(muta_project.time_signatures)
        self.first_bar_length = int(time_signatures[0].bar_length())
        tempos = self.parse_tempos(muta_project.tempos)
        singing_tracks = self.parse_singing_tracks(
            [track for track in muta_project.tracks if track.track_type == MutaTrackType.SONG]
        )
        instrumental_tracks = self.parse_instrumental_tracks(
            [track for track in muta_project.tracks if track.track_type == MutaTrackType.AUDIO]
        )
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=singing_tracks + instrumental_tracks,
        )

    def parse_time_signatures(
        self, muta_time_signatures: list[MutaTimeSignature]
    ) -> list[TimeSignature]:
        return skip_beat_list(
            [
                TimeSignature(
                    bar_index=muta_time_signature.measure_position + 1,
                    numerator=muta_time_signature.numerator,
                    denominator=muta_time_signature.denominator,
                )
                for muta_time_signature in muta_time_signatures
            ],
            0,
        )

    def parse_tempos(self, muta_tempos: list[MutaTempo]) -> list[SongTempo]:
        return [
            SongTempo(
                position=muta_tempo.position + self.first_bar_length,
                bpm=muta_tempo.bpm / 100,
            )
            for muta_tempo in muta_tempos
        ]

    def parse_singing_tracks(self, muta_singing_tracks: list[MutaTrack]) -> list[SingingTrack]:
        track_list: list[SingingTrack] = []
        for muta_track in muta_singing_tracks:
            if muta_track.song_track_data is None:
                continue
            for part in muta_track.song_track_data:
                singing_track = SingingTrack(
                    title=muta_track.name,
                    ai_singer_name="".join(chr(char) for char in part.singer_name).rstrip("\0"),
                    mute=muta_track.mute,
                    solo=muta_track.solo,
                    note_list=self.parse_notes(part.notes, part.start),
                )
                if self.options.import_pitch and (
                    pitch := self.parse_pitch(part.params.pitch_data, part.start)
                ):
                    singing_track.edited_params.pitch = pitch
                track_list.append(singing_track)
        return track_list

    def parse_pitch(self, muta_pitch: list[MutaPoint], tick_offset: int) -> ParamCurve | None:
        pitch_points = [Point.start_point()]
        pitch_points.extend(
            Point(
                x=muta_point.time + tick_offset + self.first_bar_length,
                y=(muta_point.value + 1200) if 0 < muta_point.value < 12900 else -100,
            )
            for muta_point in muta_pitch
        )
        pitch_points.append(Point.end_point())
        if len(pitch_points) > 2:
            return ParamCurve(
                points=Points(root=pitch_points),
            )

    def parse_instrumental_tracks(self, muta_tracks: list[MutaTrack]) -> list[InstrumentalTrack]:
        track_list = []
        if self.options.import_instrumental_track:
            for muta_track in muta_tracks:
                if muta_track.audio_track_data is None:
                    continue
                for part in muta_track.audio_track_data:
                    instrumental_track = InstrumentalTrack(
                        title=muta_track.name,
                        mute=muta_track.mute,
                        solo=muta_track.solo,
                        offset=part.start,
                        audio_file_path=part.file_path.rstrip("\0"),
                    )
                    track_list.append(instrumental_track)
        return track_list

    def parse_notes(self, muta_notes: list[MutaNote], tick_offset: int) -> list[Note]:
        notes = []
        for muta_note in muta_notes:
            note = Note(
                start_pos=muta_note.start + tick_offset,
                length=muta_note.length,
                key_number=139 - muta_note.key,
                lyric="".join(chr(char) for char in muta_note.lyric).rstrip("\0"),
                pronunciation=muta_note.phoneme,
            )
            notes.append(note)
        return notes



from PySide6.QtCore import QCoreApplication, QEvent, QObject, Qt, Slot
from PySide6.QtGui import QMouseEvent, QWindow
from PySide6.QtQml import QmlElement
from PySide6.QtQuick import QQuickWindow

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

QML_IMPORT_NAME = "FramelessWindow"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


@QmlElement
class FramelessWindow(QQuickWindow):
    def __init__(self, parent: QWindow | None = None, border_width: int = 5) -> None:
        super().__init__(parent)
        self.flags: Qt.WindowType = (
            self.flags | Qt.WindowType.FramelessWindowHint | Qt.WindowType.Window
        )
        self.border_width = border_width
        QCoreApplication.instance().install_event_filter(self)

    @Slot()
    def click_maximize_btn(self) -> None:
        if self.visibility != QQuickWindow.Visibility.Maximized:
            self.show_maximized()
        else:
            self.show_normal()

    def event_filter(self, obj: QObject, event: QEvent) -> bool:
        et = event.type()
        if et not in [
            QMouseEvent.Type.MouseButtonPress,
            QMouseEvent.Type.MouseMove,
        ]:
            return False

        edges = Qt.Edge(0)
        pos = event.global_pos() - self.position()
        if pos.x() < self.border_width:
            edges |= Qt.Edge.LeftEdge
        if pos.x() >= self.width - self.border_width:
            edges |= Qt.Edge.RightEdge
        if pos.y() < self.border_width:
            edges |= Qt.Edge.TopEdge
        if pos.y() >= self.height - self.border_width:
            edges |= Qt.Edge.BottomEdge

        if et == QMouseEvent.Type.MouseMove and self.window_state() == Qt.WindowState.WindowNoState:
            if edges in (
                Qt.Edge.LeftEdge | Qt.Edge.TopEdge,
                Qt.Edge.RightEdge | Qt.Edge.BottomEdge,
            ):
                self.set_cursor(Qt.CursorShape.SizeFDiagCursor)
            elif edges in (
                Qt.Edge.RightEdge | Qt.Edge.TopEdge,
                Qt.Edge.LeftEdge | Qt.Edge.BottomEdge,
            ):
                self.set_cursor(Qt.CursorShape.SizeBDiagCursor)
            elif edges in (Qt.Edge.TopEdge, Qt.Edge.BottomEdge):
                self.set_cursor(Qt.CursorShape.SizeVerCursor)
            elif edges in (Qt.Edge.LeftEdge, Qt.Edge.RightEdge):
                self.set_cursor(Qt.CursorShape.SizeHorCursor)
            else:
                self.set_cursor(Qt.CursorShape.ArrowCursor)

        elif obj == self and et == QMouseEvent.Type.MouseButtonPress and edges:
            self.start_system_resize(edges)

        return False



import dataclasses
import datetime

from wanakana import PROLONGED_SOUND_MARK
from xsdata.models.datatype import XmlTime

from libresvip.core.constants import KEY_IN_OCTAVE
from libresvip.core.lyric_phoneme.japanese import is_kana, is_romaji
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.text import uuid_str
from libresvip.utils.translation import gettext_lazy as _

from .cevio_pitch import generate_for_cevio
from .constants import DEFAULT_PHONEME, OCTAVE_OFFSET, TICK_RATE
from .model import (
    CeVIOBeat,
    CeVIOCreativeStudioProject,
    CeVIOData,
    CeVIOGroup,
    CeVIONote,
    CeVIOParameter,
    CeVIOSong,
    CeVIOSound,
    CeVIOSoundSource,
    CeVIOTempo,
    CeVIOTime,
    CeVIOUnit,
)
from .options import OutputOptions


@dataclasses.dataclass
class CeVIOGenerator:
    options: OutputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> CeVIOCreativeStudioProject:
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        ccs_project = CeVIOCreativeStudioProject()
        ccs_project.generation.svss.sound_sources.sound_source.append(
            CeVIOSoundSource(
                name=self.options.default_singer_name,
                version=self.options.default_singer_version,
                sound_source_id=self.options.default_singer_id,
            )
        )
        scene_node = ccs_project.sequence.scene

        self.first_bar_length = int(project.time_signature_list[0].bar_length())

        default_tempos = self.generate_tempos(project.song_tempo_list)
        default_beats = self.generate_time_signatures(project.time_signature_list)

        for new_unit, new_group in self.generate_tracks(
            project.track_list,
            default_tempos,
            default_beats,
            project.song_tempo_list,
        ):
            scene_node.units.unit.append(new_unit)
            scene_node.groups.group.append(new_group)

        return ccs_project

    def generate_tempos(self, tempos: list[SongTempo]) -> CeVIOTempo:
        tempo_node = CeVIOTempo(sound=[CeVIOSound(clock=0, tempo=tempos[0].bpm)])
        for model in tempos[1:]:
            tempo_node.sound.append(
                CeVIOSound(
                    clock=int(model.position * TICK_RATE),
                    tempo=model.bpm,
                )
            )
        return tempo_node

    def generate_time_signatures(self, time_signatures: list[TimeSignature]) -> CeVIOBeat:
        beat = CeVIOBeat(
            time=[
                CeVIOTime(
                    clock=0,
                    beats=time_signatures[0].numerator,
                    beat_type=time_signatures[0].denominator,
                )
            ]
        )
        tick = 0.0
        prev_time_signature = time_signatures[0]
        for time_signature in time_signatures[1:]:
            if time_signature.bar_index > prev_time_signature.bar_index:
                tick += (
                    time_signature.bar_index - prev_time_signature.bar_index
                ) * prev_time_signature.bar_length()
            beat.time.append(
                CeVIOTime(
                    clock=int(tick * TICK_RATE),
                    beats=time_signature.numerator,
                    beat_type=time_signature.denominator,
                )
            )
            prev_time_signature = time_signature
        return beat

    def generate_tracks(
        self,
        tracks: list[Track],
        tempo: CeVIOTempo,
        beat: CeVIOBeat,
        tempo_list: list[SongTempo],
    ) -> list[tuple[CeVIOUnit, CeVIOGroup]]:
        results = []
        for track in tracks:
            if isinstance(track, SingingTrack):
                new_group = CeVIOGroup(
                    group_id=uuid_str(),
                    name=track.title,
                    is_muted=track.mute,
                    is_solo=track.solo,
                    cast_id=self.options.default_singer_id,
                )
                new_unit = CeVIOUnit(
                    cast_id=self.options.default_singer_id,
                    group=new_group.group_id,
                    start_time=XmlTime.from_string("00:00:00.000"),
                    song=CeVIOSong(tempo=tempo, beat=beat),
                )
                new_unit.song.score.note = self.generate_notes(track.note_list)
                if len(track.note_list):
                    max_tick = track.note_list[-1].end_pos
                    max_secs = self.time_synchronizer.get_actual_secs_from_ticks(max_tick)
                    new_unit.duration = XmlTime.from_time(
                        datetime.datetime.fromtimestamp(max_secs, tz=datetime.timezone.utc).time()
                    )

                if log_f0 := self.generate_pitch(track.edited_params.pitch, tempo_list):
                    new_unit.song.parameter.log_f0 = log_f0
                results.append((new_unit, new_group))
            else:
                if (
                    track_info := audio_track_info(track.audio_file_path, only_wav=True)
                ) is not None:
                    start_secs = self.time_synchronizer.get_actual_secs_from_ticks(track.offset)
                    start_time = XmlTime.from_time(
                        datetime.datetime.fromtimestamp(start_secs, tz=datetime.timezone.utc).time()
                    )
                    end_time = XmlTime.from_time(
                        datetime.datetime.fromtimestamp(
                            track_info.duration / 1000,
                            tz=datetime.timezone.utc,
                        ).time()
                    )
                    new_group = CeVIOGroup(
                        group_id=uuid_str(),
                        name=track.title,
                        is_muted=track.mute,
                        is_solo=track.solo,
                        category="OuterAudio",
                    )
                    new_unit = CeVIOUnit(
                        category="OuterAudio",
                        group=new_group.group_id,
                        file_path=track.audio_file_path,
                        start_time=start_time,
                        duration=end_time,
                    )
                    results.append((new_unit, new_group))
        return results

    def generate_notes(self, notes: list[Note]) -> list[CeVIONote]:
        cevio_notes = []
        for note in notes:
            lyric = chr(PROLONGED_SOUND_MARK) if note.lyric == "-" else note.lyric
            phonetic = None
            if note.pronunciation:
                phonetic = note.pronunciation
            elif not is_kana(lyric) and not is_romaji(lyric):
                phonetic = DEFAULT_PHONEME
                msg_prefix = _("Unsupported lyric: ")
                show_warning(f"{msg_prefix} {lyric}")
            cevio_notes.append(
                CeVIONote(
                    clock=int(note.start_pos * TICK_RATE),
                    duration=int(note.length * TICK_RATE),
                    lyric=lyric,
                    phonetic=phonetic,
                    pitch_octave=note.key_number // KEY_IN_OCTAVE + OCTAVE_OFFSET,
                    pitch_step=note.key_number % KEY_IN_OCTAVE,
                )
            )
        return cevio_notes

    def generate_pitch(
        self, pitch: ParamCurve, tempo_list: list[SongTempo]
    ) -> CeVIOParameter | None:
        if (data := generate_for_cevio(pitch, tempo_list, self.first_bar_length)) is not None:
            log_f0_node = CeVIOParameter(length=data.length)
            log_f0_node.data.extend(
                CeVIOData(
                    index=each.idx,
                    repeat=each.repeat,
                    value=each.value,
                )
                for each in data.events
            )
            return log_f0_node



from libresvip.model.base import ParamCurve, Points
from libresvip.model.point import Point
from libresvip.utils.music_math import midi2hz

from .models.ds_param_curve import DsParamCurve
from .models.ds_param_node import DsParamNode


class PitchParamUtils:
    @classmethod
    def encode(cls, curve: ParamCurve, end: int, time_step: float = 0.005) -> DsParamCurve:
        end += 1920
        return DsParamCurve(
            step_size=time_step,
            point_list=cls.encode_point_list(curve.points, end),
        )

    @classmethod
    def encode_point_list(cls, os_point_list: Points, end: int) -> list[DsParamNode] | None:
        if valid_points := [
            p for p in os_point_list.root if p.x >= 1930 and p.x + 10 < end and p.y >= 0
        ]:
            return [
                DsParamNode(
                    time=pos / 1000.0,
                    value=midi2hz(cls.value_at(valid_points, pos) / 100.0),
                )
                for pos in range(1920, end, 5)
            ]
        else:
            return None

    @staticmethod
    def value_at(segment: list[Point], ticks: float) -> float:
        left_point = next((p for p in reversed(segment) if p.x <= ticks), None)
        if left_point is None:
            return segment[0].y

        right_point = next((p for p in segment if p.x > ticks), None)
        if right_point is None:
            return segment[-1].y

        ratio = (ticks - left_point.x) / (right_point.x - left_point.x)
        return (1 - ratio) * left_point.y + ratio * right_point.y



import dataclasses
import math

from libresvip.core.tick_counter import skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.music_math import linear_interpolation, midi2note

from .model import (
    PIT,
    VocalSharpBeat,
    VocalSharpMonoTrack,
    VocalSharpNote,
    VocalSharpNoteTrack,
    VocalSharpParameter,
    VocalSharpProject,
    VocalSharpSequence,
    VocalSharpStereoTrack,
    VocalSharpSyllable,
    VocalSharpTempo,
)
from .options import OutputOptions
from .vspx_interval_dict import BasePitchCurve


@dataclasses.dataclass
class VocalSharpGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> VocalSharpProject:
        vspx_project = VocalSharpProject()
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        vspx_project.project.tempo = self.generate_tempos(project.song_tempo_list)
        vspx_project.project.beat = self.generate_time_signatures(project.time_signature_list)
        singing_tracks = self.generate_singing_tracks(
            [track for track in project.track_list if isinstance(track, SingingTrack)]
        )
        vspx_project.project.tracks.extend(singing_tracks)
        vspx_project.project.tracks.extend(
            self.generate_instrumental_tracks(
                [track for track in project.track_list if isinstance(track, InstrumentalTrack)]
            )
        )
        max_duration = max(
            (note.pos + note.duration for track in singing_tracks for note in track.note),
            default=0,
        )
        vspx_duration = math.ceil(max(max_duration - 122880, 0) / 30720) * 30720 + 122880
        vspx_project.project.duration = vspx_duration
        return vspx_project

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature]
    ) -> list[VocalSharpBeat]:
        return [
            VocalSharpBeat(
                bar_index=time_signature.bar_index,
                beat_per_bar=time_signature.numerator,
                bar_divide=time_signature.denominator,
            )
            for time_signature in time_signatures
        ]

    def generate_tempos(self, tempos: list[SongTempo]) -> list[VocalSharpTempo]:
        tempos = skip_tempo_list(tempos, self.first_bar_length)
        return [
            VocalSharpTempo(
                pos=tempo.position,
                bpm=tempo.bpm,
            )
            for tempo in tempos
        ]

    def generate_instrumental_tracks(
        self, instrumental_tracks: list[InstrumentalTrack]
    ) -> list[VocalSharpMonoTrack | VocalSharpStereoTrack]:
        track_list: list[VocalSharpMonoTrack | VocalSharpStereoTrack] = []
        for track in instrumental_tracks:
            if (track_info := audio_track_info(track.audio_file_path, only_wav=True)) is not None:
                sequence = VocalSharpSequence(
                    name=track.title,
                    path=track.audio_file_path,
                    pos=track.offset,
                )
                if track_info.channel_s == 1:
                    track_list.append(
                        VocalSharpMonoTrack(
                            name=track.title,
                            is_mute=str(track.mute),
                            is_solo=str(track.solo),
                            sequences=[sequence],
                        )
                    )
                elif track_info.channel_s == 2:
                    track_list.append(
                        VocalSharpStereoTrack(
                            name=track.title,
                            is_mute=str(track.mute),
                            is_solo=str(track.solo),
                            sequences=[sequence],
                        )
                    )
        return track_list

    def generate_singing_tracks(
        self, singing_tracks: list[SingingTrack]
    ) -> list[VocalSharpNoteTrack]:
        note_tracks = []
        for track in singing_tracks:
            note_track = VocalSharpNoteTrack(
                name=track.title,
                singer=track.ai_singer_name,
                is_mute=str(track.mute),
                is_solo=str(track.solo),
                note=self.generate_notes(track.note_list),
            )
            if pitch_points := self.generate_pitch(track.edited_params.pitch, note_track):
                note_track.parameter = VocalSharpParameter(points=pitch_points)
            note_tracks.append(note_track)
        return note_tracks

    def generate_notes(self, notes: list[Note]) -> list[VocalSharpNote]:
        return [
            VocalSharpNote(
                pos=note.start_pos,
                pitch=midi2note(note.key_number),
                duration=note.length,
                lyric=note.lyric or note.pronunciation,
                syllable=[VocalSharpSyllable()],
            )
            for note in notes
        ]

    def generate_pitch(self, pitch: ParamCurve, note_track: VocalSharpNoteTrack) -> list[PIT]:
        base_pitch_curve = BasePitchCurve(note_track, None, self.synchronizer)
        pitch_points: list[PIT] = []
        prev_point: PIT | None = None
        for point in pitch.points.root:
            cur_tick = point.x - self.first_bar_length
            cur_secs = self.synchronizer.get_actual_secs_from_ticks(cur_tick)
            if (
                point.y > 0
                and (base_key := base_pitch_curve.semitone_value_at(cur_secs)) is not None
            ):
                if prev_point is not None:
                    cur_value = point.y - base_key * 100
                    pitch_points.extend(
                        PIT(
                            time=i,
                            value=round(
                                linear_interpolation(
                                    i,
                                    (prev_point.time, prev_point.value),
                                    (cur_tick, cur_value),
                                )
                            ),
                        )
                        for i in range(prev_point.time + 1, cur_tick)
                    )
                else:
                    pitch_points.append(
                        PIT(
                            time=cur_tick,
                            value=round(point.y - base_key * 100),
                        )
                    )
                prev_point = pitch_points[-1]
            else:
                prev_point = None
        return pitch_points
